from collections import defaultdict
from dataclasses import dataclass
from logging import getLogger
from typing import Awaitable, Type

from cats.identity import Identity, IdentityObject
from cats.plugins import Form, Scheme
from cats.types import Json, T_Headers
from cats.v2.action import Action, ActionLike, InputAction
from cats.v2.server import Connection

__all__ = [
    'HandlerItem',
    'Api',
    'Handler',
]

logging = getLogger('CATS')


@dataclass
class HandlerItem(object):
    id: int
    name: str
    handler: Type['Handler']
    version: int | None = None
    end_version: int | None = None


class Api:
    __slots__ = ('_handlers',)

    def __init__(self):
        self._handlers: dict[int, list[HandlerItem]] = defaultdict(list)

    def register(self, handler: HandlerItem) -> None: ...

    @property
    def handlers(self) -> dict[int, list[HandlerItem]]: ...

    def update(self, app: 'Api') -> None: ...

    def compute(self) -> dict[int, HandlerItem | list[HandlerItem]]: ...


class Handler:
    __slots__ = ('action',)
    handler_id: int

    Loader: Scheme | None = None
    Dumper: Scheme | None = None

    data_type: int | tuple[int] | None = None
    min_data_len: int | None = None
    max_data_len: int | None = None
    block_models: tuple[str] | None = None
    require_models: tuple[str] | None = None
    require_auth: bool | None = None
    min_file_size: int | None = None
    max_file_size: int | None = None
    min_file_total_size: int | None = None
    max_file_total_size: int | None = None
    min_file_amount: int | None = None
    max_file_amount: int | None = None
    fix_exec_time: int | float | None = None

    def __init__(self, action: Action):
        self.action: Action = action

    def __init_subclass__(cls, /, api: Api = None, id: int = None,
                          name: str = None, version: int = None, end_version: int = None): ...

    @property
    def conn(self) -> Connection:
        return self.action.conn

    async def __call__(self) -> ActionLike | None: ...

    async def prepare(self) -> None: ...

    async def handle(self):
        raise NotImplementedError

    async def json_load(self, *, many: bool = False, plain: bool = False) -> Json | Form: ...

    async def json_dump(self, data, *, headers: T_Headers = None,
                        status: int = 200, many: bool = None, plain: bool = False) -> ActionLike: ...

    @property
    def identity(self) -> Identity | IdentityObject | None: ...

    def ask(self, data=None, data_type: int = None, compression: int = None, *,
            headers: T_Headers = None, status: int = 200,
            bypass_limit=False, bypass_count=False, timeout=None) -> Awaitable['InputAction']: ...

    def _check_before_recv(self) -> None: ...

    def _check_after_recv(self) -> None: ...

    def _check_data_type(self) -> None: ...

    def _check_data_len(self) -> None: ...

    def _check_models(self) -> None: ...

    def _check_auth(self) -> None: ...

    def _check_file_size(self) -> None: ...

    def _check_file_total_size(self) -> None: ...

    def _check_file_amount(self) -> None: ...

    def _check_min_max(self, size: int, min_size: int | None, max_size: int | None, part: str) -> None: ...



import asyncio
from contextlib import asynccontextmanager
from logging import Logger
from traceback import format_tb
from typing import TypeVar

import rollbar
from tornado.iostream import IOStream

from cats.errors import ProtocolError
from cats.identity import Identity
from cats.types import BytesAnyGen
from cats.v2 import C_NONE, Compressor
from cats.v2.action import BaseAction, Input
from cats.v2.config import Config

__all__ = [
    'ConnType',
    'Connection',
]

ConnType = TypeVar('ConnType', bound='Connection')


class Connection:
    """
    Base connection interface, that used by both client and server side
    """
    __slots__ = (
        'download_speed',
        'input_pool',
        'send_queue',
        'send_task',
        'recv_future',
        'locker',
        'conf',
        'allowed_compressors',
        'default_compressor',
        '_closed',
        '_loop',
        '_identity',
        '_identity_timer',
        '_credentials',
        '_message_pool',
    )

    PASS_EXCEPTIONS = (
        KeyboardInterrupt,
        asyncio.CancelledError,
        asyncio.TimeoutError,
        asyncio.InvalidStateError,
    )

    address: tuple[str, int]
    _stream: IOStream
    logging: Logger

    def __init__(self, conf: Config):
        self.conf: Config = conf
        self.download_speed: int = 0
        self.input_pool: dict[int, Input] = {}
        self.send_queue: asyncio.Queue = asyncio.Queue()
        self.send_task: asyncio.Task | None = None
        self.recv_future: asyncio.Future | None = None
        self.locker: asyncio.Future | None = None
        self.allowed_compressors: set[int] = {C_NONE}
        self.default_compressor: int = C_NONE
        self._closed: bool = False
        self._loop = asyncio.get_running_loop()
        self._identity: Identity | None = None
        self._identity_timer: asyncio.TimerHandle | None = None
        self._credentials = None
        self._message_pool: list[int] = []

    def set_compressors(self, allowed: list[str], default: str = None):
        allowed = [a.lower() for a in allowed]
        self.allowed_compressors = {Compressor.codes[a] for a in allowed if a in Compressor.codes}
        self.allowed_compressors.add(C_NONE)
        if default is None or default.lower() not in Compressor.codes:
            self.default_compressor = C_NONE
        else:
            self.default_compressor = Compressor.codes[default.lower()]

    @property
    def is_open(self):
        return not self._closed and not self._stream.closed()

    @property
    def host(self) -> str:
        return self.address[0]

    @property
    def port(self) -> int:
        return self.address[1]

    async def start(self) -> None:
        """
        Init connection and start the reading loop
        :return:
        """
        self.send_task = self._loop.create_task(self.send_loop())
        self.send_task.add_done_callback(self.on_tick_done)

        await self.recv_loop()

    async def init(self):
        """Init connection state"""
        raise NotImplementedError

    async def send_loop(self):
        while self.is_open:
            waiter, self.locker = await self.send_queue.get()
            waiter.set_result(None)
            await self.locker

    async def recv_loop(self):
        while self.is_open:
            if self.recv_future is not None:
                await self.recv_future

            self.recv_future = asyncio.Future()
            task = self._loop.create_task(self.tick())
            task.add_done_callback(self.on_tick_done)

    async def tick(self):
        action_type_id = await self.read(1)
        action_class = BaseAction.get_class_by_type_id(action_type_id)
        if action_class is None:
            raise ProtocolError(f'Received unknown Action Type ID [{action_type_id.hex()}]', conn=self)

        try:
            await self.handle(await action_class.init(self))
        except self.conf.ignore_errors as err:
            self.close(err)
        except Exception as err:
            rollbar.report_exc_info(extra_data=self.identity_scope_user)  # FIXME: Use plugins/middleware for this
            self.close(err)

    def on_tick_done(self, task):
        try:
            if exc := task.exception():
                self.close(exc)
        except (asyncio.CancelledError, asyncio.InvalidStateError) as err:
            self.close(err)

    async def handle(self, action: BaseAction):
        """
        Do whatever you need with incoming actions
        :param action:
        :return:
        """
        raise NotImplementedError

    async def send(self, handler_id: int, data=None, message_id=None, compression=None, *,
                   headers=None, status=None):
        raise NotImplementedError

    async def send_stream(self, handler_id: int, data: BytesAnyGen, data_type: int,
                          message_id=None, compression=None, *,
                          headers=None, status=None):
        raise NotImplementedError

    async def read(self, num_bytes: int, partial: bool = False) -> bytes:
        """
        Should read from stream, may log something, reset timers, etc.
        :param num_bytes:
        :param partial:
        :return:
        """
        return await self._stream.read_bytes(num_bytes, partial=partial)

    async def read_until(self, delimiter: bytes, max_bytes: int | None) -> bytes:
        """
        Should read from stream, may log something, reset timers, etc.
        :param delimiter:
        :param max_bytes:
        :return:
        """
        return await self._stream.read_until(delimiter, max_bytes=max_bytes)

    async def write(self, data: bytes | bytearray | memoryview):
        """
        Should write to stream, may log something, reset timers, etc.
        :param data:
        :return:
        """
        return await self._stream.write(data)

    def get_free_message_id(self) -> int:
        raise NotImplementedError

    @asynccontextmanager
    async def preserve_message_id(self, message_id: int):
        """
        Lock message_id from being used
        e.g. next preserve_message_id() with the same message_id won't run until previous stop
        """
        try:
            while message_id in self._message_pool:
                await asyncio.sleep(0.5)
            self._message_pool.append(message_id)
            yield
        finally:
            try:
                self._message_pool.remove(message_id)
            except ValueError:
                pass

    @asynccontextmanager
    async def lock_write(self):
        """
        Lock write ability until previously called are done
        """
        waiter = asyncio.Future()
        locker = asyncio.Future()
        await self.send_queue.put((waiter, locker))
        await waiter
        yield
        locker.set_result(None)

    @property
    def identity(self) -> Identity | None:
        return self._identity

    @property
    def credentials(self):
        return self._credentials

    @property
    def identity_scope_user(self):
        if not self.signed_in:
            return {'ip': self.host}

        return self.identity.sentry_scope

    @property
    def signed_in(self) -> bool:
        return self._identity is not None

    def set_identity(self, identity: Identity, credentials=None, timeout: int | float | None = None) -> None:
        """
        Mark connection as Signed In with identity [and credentials used]
        :param identity: Identity subclass object
        :param credentials: anything that was used to get identity, may be JWT, dict, key, etc.
        :param timeout: How long in seconds should identity be stored. 0 | None for always
        :return:
        """
        self._identity: Identity | None = identity
        self._credentials = credentials

        self.prolong_identity(timeout)

        self.debug(f'Signed in as {self.identity_scope_user} <{self.host}:{self.port}>')

    def prolong_identity(self, timeout: int | float | None = None) -> None:
        """
        Recreates identity removal timer with new timeout. If timeout is None, no timer is created
        :param timeout:
        :return: How long in seconds should identity be stored. 0 | None for always
        """
        if self._identity_timer is not None:
            self._identity_timer.cancel()
            self._identity_timer = None
        if timeout:
            self._identity_timer = self._loop.call_later(timeout, self.sign_out)

    def sign_out(self) -> None:
        """
        Mark connection as Signed Out
        :return:
        """
        if not self.signed_in:
            return
        self.prolong_identity(None)
        self._identity = None
        self._credentials = None
        self.debug(f'Signed out from {self.identity_scope_user} <{self.host}:{self.port}>')

    def close(self, exc: BaseException = None) -> None:
        """
        Close stream if opened, log exceptions, clear caches, etc.
        :param exc: Exception that caused closing
        :return:
        """
        if self._closed:
            return

        self._closed = True
        self.sign_out()
        if exc and not isinstance(exc, self.conf.ignore_errors):
            self.logging.error(f'{exc.__class__.__qualname__} {exc}\n{format_tb(exc.__traceback__)}')
        self._stream.close(exc)
        for task in self._close_tasks():
            if task is None:
                continue
            if isinstance(task, asyncio.TimerHandle):
                task.cancel()
            elif not task.done():
                if exc and not isinstance(task, asyncio.Task):
                    task.set_exception(exc)
                else:
                    task.cancel()

    def _close_tasks(self):
        yield self.locker
        yield self.recv_future
        yield self.send_task

    def debug(self, msg: str, *args, **kwargs):
        if self.conf.debug:
            self.logging.debug(msg, *args, **kwargs)

    async def __aenter__(self) -> 'ConnType':
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close(exc_val)



from typing import Type, TypeVar

__all__ = [
    'IdentityObject',
    'Identity',
]

IdentityObject = TypeVar('IdentityObject', bound='Identity')


class Identity:
    __slots__ = ()
    identity_list: list[Type['Identity']] = []

    @property
    def sentry_scope(self) -> dict:
        """
        Must return dict of data that can be used by Sentry
        """
        raise NotImplementedError

    def __init_subclass__(cls):
        cls.identity_list.append(cls)



import asyncio
from io import BytesIO
from pathlib import Path
from struct import Struct
from time import time_ns
from typing import NamedTuple, Type, TypeAlias, TypeVar

import math
import struct_model

from cats.errors import CatsUsageError, InputCancelled, MalformedDataError, ProtocolError
from cats.types import Bytes, Headers
from cats.utils import Delay, as_uint, bytes2hex, filter_json, format_amount, int2hex, tmp_file, to_uint
from cats.v2.codecs import Codec, T_FILE, T_JSON
from cats.v2.compression import Compressor

__all__ = [
    'ActionLike',
    'Input',
    'BaseAction',
    'BasicAction',
    'Action',
    'StreamAction',
    'InputAction',
    'DownloadSpeedAction',
    'CancelInputAction',
    'PingAction',
]

MAX_IN_MEMORY = 1 << 24
MAX_CHUNK_READ = 1 << 20
PROPOSAL_PLACEHOLDER = bytes(5000)

ActionLike: TypeAlias = TypeVar('ActionLike', bound='Action')


class Input:
    __slots__ = ('future', 'conn', 'message_id', 'bypass_count', 'timer')

    def __init__(self,
                 future,
                 conn,
                 message_id,
                 bypass_count=False,
                 timeout=None):
        self.timer = None
        self.future = future
        self.conn = conn
        self.message_id = message_id
        self.bypass_count = bypass_count
        if timeout:
            self.timer = asyncio.get_running_loop().call_later(timeout, self.cancel)

    def done(self, result):
        self.future.set_result(result)
        self.cancel()

    def cancel(self):
        if self.timer is not None:
            self.timer.cancel()
            self.timer = None
        if not self.future.done():
            self.future.cancel()
        self.conn.input_pool.pop(self.message_id, None)


class BaseAction(dict):
    __slots__ = ('data', 'headers', 'message_id', 'conn')
    __registry__ = {}
    HEADER_SEPARATOR = b'\x00\x00'

    type_id: bytes
    head_struct: Struct
    head_tuple: Type[NamedTuple]

    def __init__(self, data=None, *, headers=None, status=None,
                 message_id=None):
        assert type(self) is not BaseAction, 'Creation of BaseAction instances are prohibited'

        if headers is not None and not isinstance(headers, dict):
            raise MalformedDataError('Invalid Headers provided', data=data, headers=headers)

        self.data = data
        self.headers = Headers(headers or {})
        self.status = status or self.status
        self.message_id = message_id
        self.conn = None
        super().__init__()

    def __init_subclass__(cls, *, abstract=False):
        if abstract:
            return
        assert cls.type_id not in cls.__registry__, f'ActionType with ID {cls.type_id} already assigned'
        cls.__registry__[cls.type_id] = cls

    @property
    def status(self):
        return self.headers.get('Status', 200)

    @status.setter
    def status(self, value=None):
        if value is None:
            value = 200
        elif not isinstance(value, int):
            raise TypeError('Invalid status type')
        self.headers['Status'] = value

    @status.deleter
    def status(self):
        self.headers['Status'] = 200

    @property
    def offset(self) -> int:
        return self.headers.get('Offset', 0)

    @offset.setter
    def offset(self, value: int = None):
        if value is None:
            value = 0
        elif not isinstance(value, int):
            raise TypeError('Invalid offset type')
        self.headers['Offset'] = value

    @offset.deleter
    def offset(self):
        self.headers['Offset'] = 0

    @classmethod
    def get_class_by_type_id(cls, type_id):
        return cls.__registry__.get(type_id)

    async def ask(self, data=None, data_type=None, compression=None, *,
                  headers=None, status=None,
                  bypass_limit=False, bypass_count=False,
                  timeout=None):
        if not self.conn:
            raise CatsUsageError('Connection is not set')
        fut = asyncio.Future()
        timeout = self.conn.conf.input_timeout if timeout is None else timeout

        if not bypass_limit:
            amount = sum(not i.bypass_count for i in self.conn.input_pool.values())
            if amount > self.conn.conf.input_limit:
                k = min(self.conn.input_pool.keys())
                self.conn.input_pool[k].cancel()

        inp = Input(fut, self.conn, self.message_id, bypass_count, timeout)
        if self.message_id in self.conn.input_pool:
            raise ProtocolError(f'Input query with MID {self.message_id} already exists', conn=self.conn)

        self.conn.input_pool[self.message_id] = inp
        action = InputAction(data, headers=headers, status=status, message_id=self.message_id,
                             data_type=data_type, compression=compression)
        await action.send(self.conn)
        try:
            return await fut
        except (asyncio.TimeoutError, asyncio.CancelledError) as err:
            raise InputCancelled from err

    @classmethod
    async def init(cls, conn):
        raise NotImplementedError

    @classmethod
    async def _recv_head(cls, conn):
        raise NotImplementedError

    async def dump_data(self, size: int) -> None:
        while size > 0:
            size -= len(await self.conn.read(min(size, MAX_CHUNK_READ), partial=True))
        if fut := self.conn.recv_future:
            fut.set_result(None)

    async def send(self, conn):
        raise NotImplementedError

    def __repr__(self):
        return f'{type(self).__name__}(data={str(self.data)[:256]}, headers={self.headers}, ' \
               f'status={self.status}, message_id={self.message_id})'


class BasicAction(BaseAction, abstract=True):
    __slots__ = ('data_len', 'data_type', 'compression', 'encoded')

    def __init__(self, data=None, *, headers=None, status=None,
                 message_id=None, data_len=None, data_type=None, compression=None, encoded=None):
        self.data_len = data_len
        self.data_type = data_type
        self.compression = compression
        self.encoded = encoded
        super().__init__(data, headers=headers, status=status, message_id=message_id)

    @classmethod
    async def init(cls, conn):
        raise NotImplementedError

    @classmethod
    async def _recv_head(cls, conn):
        raise NotImplementedError

    async def send(self, conn):
        raise NotImplementedError

    async def recv_data(self):
        if self.data_len > MAX_IN_MEMORY:
            if self.data_type != T_FILE:
                raise ProtocolError(f'Attempted to send message larger than '
                                    f'{format_amount(MAX_IN_MEMORY)}', conn=self.conn)
            await self._recv_large_data()
        else:
            await self._recv_small_data()

        if fut := self.conn.recv_future:
            fut.set_result(None)

    async def _recv_small_data(self):
        left = self.data_len
        buff = bytearray()
        while left > 0:
            chunk, left = await self._recv_chunk(left)
            buff.extend(chunk)

        buff = await Compressor.decompress(buff, self.headers, compression=self.compression)
        self.data = await Codec.decode(buff, self.data_type, self.headers)
        if self.data_type == T_JSON:
            self.conn.debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {filter_json(self.data)}')

    async def _recv_large_data(self):
        left = self.data_len
        src, dst = tmp_file(), tmp_file()

        try:
            with src.open('wb') as fh:
                while left > 0:
                    chunk, left = await self._recv_chunk(left)
                    fh.write(chunk)

            await Compressor.decompress_file(src, dst, self.headers, compression=self.compression)
            self.data = await Codec.decode(dst, self.data_type, self.headers)
        except Exception:
            dst.unlink(missing_ok=True)
            raise
        finally:
            src.unlink(missing_ok=True)

    async def _recv_chunk(self, left):
        chunk = await self.conn.read(min(left, MAX_CHUNK_READ), partial=True)
        self.conn.debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
        left -= len(chunk)
        return chunk, left

    async def _encode(self):
        if self.encoded is None:
            data, data_type = await Codec.encode(self.data, self.headers, self.offset)
        else:
            data = self.data
            data_type = self.encoded
        if isinstance(data, Path):
            data, buff = tmp_file(), data
            compression = await Compressor.compress_file(buff, data, self.headers,
                                                         self.conn.allowed_compressors, self.conn.default_compressor,
                                                         self.compression)
            data_len = data.stat().st_size
        else:
            data, compression = await Compressor.compress(data, self.headers,
                                                          self.conn.allowed_compressors, self.conn.default_compressor,
                                                          self.compression)
            data_len = len(data)
        return data, data_len, data_type, compression

    async def _write_data_to_stream(self, conn, data, data_len, data_type, compression):
        if isinstance(data, Path):
            fh = data.open('rb')
            left = data.stat().st_size
        else:
            fh = BytesIO(data)
            left = len(data)

        try:
            max_chunk_size = min(MAX_IN_MEMORY, conn.download_speed) or MAX_IN_MEMORY
            delay = Delay(conn.download_speed)
            while left > 0:
                size = min(left, max_chunk_size)
                chunk = fh.read(size)
                left -= size
                await delay(size)
                conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {bytes2hex(chunk[:64])}...')
                await conn.write(chunk)
        finally:
            fh.close()
            if data_type == T_JSON:
                conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {filter_json(self.data)}')

    def __repr__(self):
        return f'{type(self).__name__}(data={str(self.data)[:256]}, headers={self.headers}, ' \
               f'status={self.status}, message_id={self.message_id}, ' \
               f'data_len={self.data_len}, data_type={self.data_type}, ' \
               f'compression={self.compression}, encoded={self.encoded})'


class Action(BasicAction):
    __slots__ = ('handler_id', 'send_time')

    type_id = b'\x00'

    class Head(struct_model.StructModel):
        handler_id: struct_model.uInt2
        message_id: struct_model.uInt2
        send_time: struct_model.uInt8
        data_type: struct_model.uInt1
        compression: struct_model.uInt1
        data_len: struct_model.uInt4

    def __init__(self, data=None, *, headers=None, status=None, message_id=None,
                 handler_id=None, data_len=None, data_type=None, compression=None,
                 send_time=None, encoded=None):
        assert compression is None or isinstance(compression, int), 'Invalid compression type'
        assert data_type is None or isinstance(data_type, int), 'Invalid data type provided'

        self.handler_id = handler_id
        self.send_time = send_time or (time_ns() // 1000_000)
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         data_len=data_len, data_type=data_type, compression=compression, encoded=encoded)

    @classmethod
    async def init(cls, conn):
        head = await cls._recv_head(conn)
        headers = await conn.read_until(cls.HEADER_SEPARATOR, head.data_len)
        head.data_len -= len(headers)
        headers = Headers.decode(headers[:-2])
        conn.debug(f'[RECV {conn.address}] [{int2hex(head.message_id):<4}] <- HEADERS {headers}')

        action = cls(**vars(head), headers=headers)
        action.conn = conn
        return action

    @classmethod
    async def _recv_head(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[RECV {conn.address}] Request  '
                   f'H: {int2hex(head.handler_id):<4} '
                   f'M: {int2hex(head.message_id):<4} '
                   f'L: {format_amount(head.data_len):<8}'
                   f'T: {Codec.codecs[head.data_type].type_name:<8} '
                   f'C: {Compressor.compressors[head.compression].type_name:<8}')
        return head

    async def send(self, conn):
        self.conn = conn
        data, data_len, data_type, compression = await self._encode()

        try:
            message_headers = self.headers.encode() + self.HEADER_SEPARATOR

            _data_len = data_len + len(message_headers)
            header = self.type_id + self.Head(
                self.handler_id,
                self.message_id,
                time_ns() // 1000_000,
                data_type,
                compression,
                _data_len
            ).pack() + message_headers

            async with conn.lock_write():
                await conn.write(header)
                conn.debug(f'[SEND {conn.address}] Response '
                           f'H: {int2hex(self.handler_id):<4} '
                           f'M: {int2hex(self.message_id):<4} '
                           f'L: {format_amount(_data_len):<8}'
                           f'T: {Codec.codecs[data_type].type_name:<8} '
                           f'C: {Compressor.compressors[compression].type_name:<8} ')
                conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
                await self._write_data_to_stream(conn, data, data_len, data_type, compression)
        finally:
            if isinstance(data, Path):
                data.unlink(missing_ok=True)

    def __repr__(self):
        return f'{type(self).__name__}(data={str(self.data)[:256]}, headers={self.headers}, ' \
               f'status={self.status}, message_id={self.message_id}, ' \
               f'data_len={self.data_len}, data_type={self.data_type}, ' \
               f'compression={self.compression}, send_time={self.send_time}, encoded={self.encoded})'


class StreamAction(Action):
    type_id = b'\x01'

    class Head(struct_model.StructModel):
        handler_id: struct_model.uInt2
        message_id: struct_model.uInt2
        send_time: struct_model.uInt8
        data_type: struct_model.uInt1
        compression: struct_model.uInt1

    def __init__(self, data=None, *, headers=None, status=None, message_id=None,
                 handler_id=None, data_type=None, compression=None,
                 send_time=None):
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         handler_id=handler_id, data_type=data_type, compression=compression,
                         send_time=send_time)

    @classmethod
    async def init(cls, conn):
        head = await cls._recv_head(conn)
        headers_size = as_uint(await conn.read(4))
        headers = Headers.decode(await conn.read(headers_size))
        conn.debug(f'[RECV {conn.address}] [{int2hex(head.message_id):<4}] <- HEADERS {headers}')

        action = cls(**vars(head), headers=headers)
        action.conn = conn
        return action

    @classmethod
    async def _recv_head(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[RECV {conn.address}] Stream   '
                   f'H: {int2hex(head.handler_id):<4} '
                   f'M: {int2hex(head.message_id):<4} '
                   f'T: {Codec.codecs[head.data_type].type_name:<8} '
                   f'C: {Compressor.compressors[head.compression].type_name:<8} ')
        return head

    async def recv_data(self):
        data_len = 0
        buff = tmp_file()
        try:
            with buff.open('wb') as fh:
                while chunk_size := as_uint(await self.conn.read(4)):
                    if chunk_size > MAX_IN_MEMORY:
                        data_len += await self._recv_large_chunk(fh, chunk_size)
                    else:
                        data_len += await self._recv_small_chunk(fh, chunk_size)

            if data_len > self.conn.conf.max_plain_payload:
                if self.data_type != T_FILE:
                    raise ProtocolError(f'Attempted to send message larger than '
                                        f'{format_amount(self.conn.conf.max_plain_payload)}', conn=self.conn)
                decode = buff
            elif self.data_type != T_FILE:
                with buff.open('rb') as _fh:
                    decode = _fh.read()
            self.data = await Codec.decode(decode, self.data_type, self.headers)
            self.data_len = data_len
        finally:
            if fut := self.conn.recv_future:
                fut.set_result(None)
            buff.unlink(missing_ok=True)

    async def _recv_large_chunk(self, fh, chunk_size):
        left = chunk_size
        part, dst = tmp_file(), tmp_file()
        try:
            with part.open('wb') as tmp:
                while left > 0:
                    chunk = await self.conn.read(min(left, MAX_CHUNK_READ), partial=True)
                    self.conn.debug(
                        f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
                    left -= len(chunk)
                    tmp.write(chunk)
            await Compressor.decompress_file(part, dst, self.headers, compression=self.compression)
            data_len = dst.stat().st_size
            with dst.open('rb') as tmp:
                while i := tmp.read(MAX_IN_MEMORY):
                    fh.write(i)
            return data_len
        finally:
            part.unlink(missing_ok=True)
            dst.unlink(missing_ok=True)

    async def _recv_small_chunk(self, fh, chunk_size):
        left = chunk_size
        part = bytearray()
        while left > 0:
            chunk = await self.conn.read(min(left, MAX_CHUNK_READ), partial=True)
            self.conn.debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
            left -= len(chunk)
            part += chunk
        part = await Compressor.decompress(part, self.headers, compression=self.compression)
        fh.write(part)
        return len(part)

    async def _encode_gen(self, conn):
        data = self.data
        compression = self.compression
        if compression is None:
            compression = await Compressor.propose_compression(PROPOSAL_PLACEHOLDER, self.headers,
                                                               conn.default_compressor)

        assert hasattr(data, '__iter__') or hasattr(data, '__aiter__'), \
            'StreamResponse payload is not (Async)Generator[Bytes, None, None]'
        data = self._async_gen(data, conn.download_speed)
        return data, compression

    async def send(self, conn):
        self.conn = conn
        data, compression = await self._encode_gen(conn)

        header = self.type_id + self.Head(
            self.handler_id,
            self.message_id,
            time_ns() // 1000_000,
            self.data_type,
            compression
        ).pack()
        message_headers = self.headers.encode()

        async with conn.lock_write():
            await conn.write(header)
            await conn.write(to_uint(len(message_headers), 4))
            await conn.write(message_headers)
            conn.debug(f'[SEND {conn.address}] Stream   '
                       f'H: {int2hex(self.handler_id):<4} '
                       f'M: {int2hex(self.message_id):<4} '
                       f'T: {Codec.codecs[self.data_type].type_name:<8} '
                       f'C: {Compressor.compressors[compression].type_name:<8} ')
            conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
            await self._write_data_to_stream(conn, data, compression=compression)

    async def _write_data_to_stream(self, conn, data, *_, compression):
        offset = self.offset
        delay = Delay(conn.download_speed)
        async for chunk in data:
            if offset > 0:
                offset -= (i := min(offset, len(chunk)))
                chunk = chunk[i:]
            if not chunk:
                continue
            if not isinstance(chunk, Bytes):
                raise MalformedDataError('Provided data chunk is not binary', data=data, headers=self.headers)

            chunk, _ = await Compressor.compress(chunk, self.headers,
                                                 conn.allowed_compressors, conn.default_compressor,
                                                 compression)
            chunk_size = len(chunk)
            if chunk_size >= 1 << 32:
                raise MalformedDataError('Provided data chunk exceeded max chunk size', data=data, headers=self.headers)

            await delay(chunk_size + 4)
            await conn.write(to_uint(chunk_size, 4))
            await conn.write(chunk)
            conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {bytes2hex(chunk[:64])}...')
        await conn.write(b'\x00\x00\x00\x00')

    async def _async_gen(self, gen, chunk_size):
        size = max(chunk_size, MAX_IN_MEMORY) or MAX_IN_MEMORY
        if hasattr(gen, '__iter__'):
            for item in gen:
                for i in range(math.ceil(len(item) / size)):
                    yield item[i * size:]
        else:
            async for item in gen:
                for i in range(math.ceil(len(item) / size)):
                    yield item[i * size:]

    def __repr__(self):
        return f'{type(self).__name__}(data={str(self.data)[:256]}, headers={self.headers}, ' \
               f'status={self.status}, message_id={self.message_id}, ' \
               f'data_type={self.data_type}, compression={self.compression}'


class InputAction(Action):
    type_id = b'\x02'

    class Head(struct_model.StructModel):
        message_id: struct_model.uInt2
        data_type: struct_model.uInt1
        compression: struct_model.uInt1
        data_len: struct_model.uInt4

    @classmethod
    async def _recv_head(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[RECV {conn.address}] Answer   '
                   f'M: {int2hex(head.message_id):<4} '
                   f'L: {format_amount(head.data_len):<8}'
                   f'T: {Codec.codecs[head.data_type].type_name:<8} '
                   f'C: {Compressor.compressors[head.compression].type_name:<8} ')
        return head

    async def send(self, conn):
        self.conn = conn
        data, data_len, data_type, compression = await self._encode()
        try:
            message_headers = self.headers.encode() + self.HEADER_SEPARATOR
            _data_len = data_len + len(message_headers)
            header = self.type_id + self.Head(
                self.message_id,
                data_type,
                compression,
                _data_len
            ).pack() + message_headers

            async with conn.lock_write():
                await conn.write(header)
                conn.debug(f'[SEND {conn.address}] Input    '
                           f'M: {int2hex(self.message_id):<4} '
                           f'L: {format_amount(_data_len):<8}'
                           f'T: {Codec.codecs[data_type].type_name:<8} '
                           f'C: {Compressor.compressors[compression].type_name:<8} ')
                conn.debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
                await self._write_data_to_stream(conn, data, data_len, data_type, compression)
        finally:
            if isinstance(data, Path):
                data.unlink(missing_ok=True)

    async def reply(self, data=None, data_type=None, compression=None, *,
                    headers=None, status=None) -> ActionLike | None:
        action = InputAction(data, headers=headers, status=status, message_id=self.message_id,
                             data_type=data_type, compression=compression)
        action.offset = self.offset
        await action.send(self.conn)
        return await self.conn.recv(self.message_id)  # noqa

    async def cancel(self) -> ActionLike | None:
        res = CancelInputAction(self.message_id)
        await res.send(self.conn)
        return await self.conn.recv(self.message_id)  # noqa


class DownloadSpeedAction(BaseAction):
    type_id = b'\x05'

    class Head(struct_model.StructModel):
        speed: struct_model.uInt4

    def __init__(self, speed):
        super().__init__()
        self.speed = speed

    @classmethod
    async def init(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[RECV {conn.address}] SET Download speed: {format_amount(head.speed)}')
        action = cls(**vars(head))
        action.conn = conn
        return action

    async def send(self, conn):
        async with conn.lock_write():
            speed: int = self.data
            await conn.write(self.type_id + to_uint(speed, 4))
            conn.debug(f'[SEND {conn.address}] SET Download speed: {format_amount(speed)}')

    def __repr__(self):
        return f'{type(self).__name__}(speed={self.speed})'


class CancelInputAction(BaseAction):
    type_id = b'\x06'

    class Head(struct_model.StructModel):
        message_id: struct_model.uInt2

    @classmethod
    async def init(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[RECV {conn.address}] CANCEL Input M: {int2hex(head.message_id):<4}')
        action = cls(**vars(head))
        action.conn = conn
        return action

    async def send(self, conn):
        async with conn.lock_write():
            message_id: int = self.data
            await conn.write(self.type_id + to_uint(message_id, 2))
            conn.debug(f'[SEND {conn.address}] CANCEL Input M: {message_id}')


class PingAction(BaseAction):
    type_id = b'\xFF'

    class Head(struct_model.StructModel):
        send_time: struct_model.uInt8

    def __init__(self, send_time=None):
        super().__init__()
        self.recv_time = time_ns() // 1000_000
        self.send_time = send_time or self.recv_time

    @classmethod
    async def init(cls, conn):
        buff = await conn.read(cls.Head.struct.size)
        head = cls.Head.unpack(buff)
        conn.debug(f'[PING {conn.address}] {head.send_time}')

        action = cls(**vars(head))
        action.conn = conn
        return action

    async def send(self, conn):
        async with conn.lock_write():
            now = time_ns() // 1000_000
            await conn.write(self.type_id + to_uint(now, 8))
            conn.debug(f'[SEND {conn.address}] PONG {now}')

    def __repr__(self):
        return f'{type(self).__name__}(send_time={self.send_time})'



import asyncio
import importlib
import math
import re
import tempfile
from logging import getLogger
from pathlib import Path
from time import time

import ujson

from cats.types import Bytes, Json

__all__ = [
    'to_uint',
    'as_uint',
    'Delay',
    'require',
    'tmp_file',
    'int2hex',
    'bytes2hex',
    'format_amount',
    'filter_json',
]

logging = getLogger('CATS.utils')


def to_uint(number: int, length: int = None) -> bytes:
    if length is None:
        length = (number.bit_length() + 7) // 8
    return number.to_bytes(length, byteorder='big', signed=False)


def as_uint(data: bytes) -> int:
    return int.from_bytes(data, byteorder='big', signed=False)


class Delay:
    """
    Rate-limiter, that calls asyncio.sleep(N), each time an object called.
    N is calculated based on specified speed, size of latest chunk and ∆time
    """

    def __init__(self, speed: int = 0):
        self.speed: int = speed
        self.start: float = time()
        self.sent: float = 0.0

    async def __call__(self, length: int = 0):
        if not self.speed:
            return

        went = time() - self.start + 0.01
        self.start = time()
        self.sent = max(0.0, length + self.sent - self.speed * went)
        if not self.sent:
            return
        await asyncio.sleep(self.sent / self.speed)


def tmp_file(**kwargs) -> Path:
    """
    Creates NamedTemporaryFile and returns associated pathlib.Path
    :param kwargs: NamedTemporaryFile arguments
    :return: pathlib.Path(temp_file)
    """
    kwargs['delete'] = kwargs.get('delete', False)
    return Path(tempfile.NamedTemporaryFile(**kwargs).name)


def require(dotted_path: str, /, *, strict: bool = True):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.

    :param dotted_path: Path to attribute/class in module
    :param strict: should the error be thrown
    :return: attribute/class
    """
    try:
        try:
            module_path, class_name = dotted_path.rsplit('.', 1)
        except ValueError as err:
            raise ImportError(f"{dotted_path} doesn't look like a module path") from err

        module = importlib.import_module(module_path)

        try:
            return getattr(module, class_name)
        except AttributeError as err:
            raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class') from err
    except ImportError:
        logging.error(f'Failed to import {dotted_path}')
        if strict:
            raise
        return None


def int2hex(number: int, chars: int = 2) -> str:
    """
    Converts integers to human-readable HEX
    :param number: integer to convert
    :param chars: minimal amount of chars(bytes) to show
    :return: HEX string w/o prefix
    """
    a = hex(number)[2:].upper()
    chars = max(len(a) // 2, chars)
    return ('00' * chars + a)[-chars * 2:]


def bytes2hex(buffer: Bytes, *, separator: str = ' ', prefix: bool = False) -> str:
    """
    Converts byte-like to HEX string

    :param buffer: what to encode
    :param separator: separator string
    :param prefix: use 0x prefix
    :return: HEX style string
    :raise TypeError:
    """
    if not isinstance(buffer, Bytes):
        raise TypeError(f'Invalid buffer type = {type(buffer)}')

    hexadecimal = buffer.hex().upper()
    parts = (hexadecimal[i: i + 2] for i in range(0, len(hexadecimal), 2))
    if prefix:
        parts = ('0x' + part for part in parts)
    return separator.join(parts)


def format_amount(num: int, *, base: int = 1024, prefix: str = '', suffix: str = 'B') -> str:
    """
    Humanize any amount to string
    format_amount(0) == "0B"
    format_amount(1750) == "1.7KB"
    format_amount(1234567890) == "1.1GB"
    format_amount(2000, base=10, suffix='r') == "2Gr"

    :param num: numeric amount
    :param base: how much does the step takes (default 1024 as for bytes)
    :param prefix: type prefix (default '')
    :param suffix: type suffix (default B as for bytes)
    :return: Formatted amount string
    :raise TypeError:
    """
    if not num:
        return f'0{suffix}'

    if not isinstance(num, int):
        raise TypeError(f'Invalid num type = {type(num)}')

    if not isinstance(base, int):
        raise TypeError(f'Invalid base type = {type(base)}')

    sign = '-' if num < 0 else ''
    num = abs(num)

    magnitude = int(math.floor(math.log(num, base)))
    val = float(num / math.pow(base, magnitude))

    if magnitude > 7:
        val = f'{val:.0f}' if val.is_integer() else f'{val:.1f}'
        return f'{sign}{val}{"Y"}{prefix}{suffix}'

    val = f'{val:.0f}' if val.is_integer() else f'{val:3.1f}'
    return f'{sign}{val}{["", "K", "M", "G", "T", "P", "E", "Z"][magnitude]}{prefix}{suffix}'


_HIDE = re.compile(r'.*(password|key|secret|jwt|pwd|пароль|ключ|секрет).*', re.IGNORECASE | re.UNICODE)


def filter_json(json: Json | Bytes, max_len: int = 16, max_size: int = 64, indent: bool = False) -> str:
    if isinstance(json, Bytes):
        json = ujson.loads(json)
    result = _filter_json_part(json, max_len, max_size)
    return ujson.dumps(result, ensure_ascii=False, escape_forward_slashes=False, indent=2 if indent else 0)


def _filter_json_part(json: Json, max_len: int = 16, max_size: int = 64) -> Json:
    if isinstance(json, dict):
        return {
            k:
                '<masked>' if (isinstance(k, str) and _HIDE.match(k))
                else v if str(k).lower() in ('error', 'exception')
                else _filter_json_part(v, max_len, max_size)
            for k, v in (*tuple(json.items())[:max_size], ('<more>', f'{len(json) - max_len} items'))[:len(json)]
        }
    if isinstance(json, list):
        return [
            _filter_json_part(v, max_len, max_size)
            for v in (*json[:max_size], f'{len(json) - max_len} more')[:len(json)]
        ]
    if isinstance(json, str) and len(json) > max_len:
        return json[:max_len] + '...'
    return json



import gzip
import shutil
from pathlib import Path

import zlib

from cats.errors import ClientSupportError, CompressorError, InvalidCompressorError
from cats.types import T_Headers
from cats.utils import as_uint, to_uint

__all__ = [
    'C_NONE',
    'C_ZLIB',
    'C_GZIP',
    'BaseCompressor',
    'DummyCompressor',
    'GzipCompressor',
    'ZlibCompressor',
    'Compressor',
]


class BaseCompressor:
    type_id: int
    type_name: str

    @classmethod
    async def compress(cls, data: bytes, headers: T_Headers) -> bytes:
        raise NotImplementedError

    @classmethod
    async def decompress(cls, data: bytes, headers: T_Headers) -> bytes:
        raise NotImplementedError

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        raise NotImplementedError

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        raise NotImplementedError


class DummyCompressor(BaseCompressor):
    type_id = 0x00
    type_name = 'dummy'

    @classmethod
    async def compress(cls, data: bytes, headers: T_Headers) -> bytes:
        return data

    @classmethod
    async def decompress(cls, data: bytes, headers: T_Headers) -> bytes:
        return data

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        shutil.copy(src.resolve().as_posix(), dst.resolve().as_posix())

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        shutil.copy(src.resolve().as_posix(), dst.resolve().as_posix())


class GzipCompressor(BaseCompressor):
    type_id = 0x01
    type_name = 'GZip'

    @classmethod
    async def compress(cls, data: bytes, headers: T_Headers) -> bytes:
        return gzip.compress(data, compresslevel=6)

    @classmethod
    async def decompress(cls, data: bytes, headers: T_Headers) -> bytes:
        return gzip.decompress(data)

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        with src.open('rb') as rc:
            with gzip.open(dst.resolve().as_posix(), 'wb', compresslevel=6) as wc:
                while line := rc.read(1 << 24):
                    wc.write(line)

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        with gzip.open(src.resolve().as_posix(), 'rb') as rc:
            with dst.open('wb') as wc:
                while line := rc.read(1 << 24):
                    wc.write(line)


class ZlibCompressor(BaseCompressor):
    """Modified ZLib compressor: uint4 length of original data will be prepended to result payload"""
    type_id = 0x02
    type_name = 'ZLib'

    @classmethod
    async def compress(cls, data: bytes, headers: T_Headers) -> bytes:
        headers['Adler32'] = zlib.adler32(data)
        return to_uint(len(data), 4) + zlib.compress(data, level=6)

    @classmethod
    async def decompress(cls, data: bytes, headers: T_Headers) -> bytes:
        ln, data = as_uint(data[:4]), data[4:]
        buff = zlib.decompress(data)
        checksum = headers.get('Adler32', None)
        if ln != len(buff):
            raise CompressorError('Broken data received: Length mismatch', data=data, headers=headers)
        if checksum is not None and zlib.adler32(buff) != checksum:
            raise CompressorError('Broken data received: Checksum mismatch', data=data, headers=headers)
        return buff

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        compressor = zlib.compressobj(level=6)
        value = 1
        ln = src.stat().st_size
        with src.open('rb') as rc:
            with dst.open('wb') as wc:
                wc.write(to_uint(ln, 4))
                while line := rc.read(1 << 24):
                    value = zlib.adler32(line, value)
                    wc.write(compressor.compress(line))
                wc.write(compressor.flush())
        del compressor
        headers['Adler32'] = value

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, headers: T_Headers) -> None:
        compressor = zlib.decompressobj()
        value = 1
        with src.open('rb') as rc:
            with dst.open('wb') as wc:
                ln = as_uint(rc.read(4))
                while line := rc.read(1 << 24):
                    wc.write(buff := compressor.decompress(line))
                    value = zlib.adler32(buff, value)
                wc.write(compressor.flush())
        if dst.stat().st_size != ln:
            raise CompressorError('Broken data received: Length mismatch', data=src, headers=headers)
        checksum = headers.get('Adler32', None)
        if checksum is not None and value != checksum:
            dst.unlink(missing_ok=True)
            raise CompressorError('Broken data received: Checksum mismatch', data=src, headers=headers)


C_NONE = DummyCompressor.type_id
C_GZIP = GzipCompressor.type_id
C_ZLIB = ZlibCompressor.type_id


class Compressor:
    codes = {
        DummyCompressor.type_name.lower(): DummyCompressor.type_id,
        GzipCompressor.type_name.lower(): GzipCompressor.type_id,
        ZlibCompressor.type_name.lower(): ZlibCompressor.type_id,
    }
    compressors = {
        C_NONE: DummyCompressor,
        C_GZIP: GzipCompressor,
        C_ZLIB: ZlibCompressor,
    }

    @classmethod
    async def compress(cls, buff: bytes, headers: T_Headers, allowed: set[int], default: int,
                       compression: int = None) -> (bytes, int):
        try:
            if compression is None:
                compression = await cls.propose_compression(buff, headers, default)
            if compression not in allowed:
                raise ClientSupportError(f'Compression unsupported by client: {cls.compressors[compression].type_name}')
            buff = await cls.compressors[compression].compress(buff, headers)
            return buff, compression

        except (KeyError, ValueError, TypeError) as err:
            raise CompressorError(f'Failed to compress data: {str(err)}', data=buff, headers=headers) from err

    @classmethod
    async def decompress(cls, buff: bytes, headers: T_Headers, compression: int) -> bytes:
        try:
            return await cls.compressors[compression].decompress(buff, headers)
        except (KeyError, ValueError, TypeError) as err:
            raise CompressorError(f'Failed to decompress data: {str(err)}', data=buff, headers=headers) from err

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, headers: T_Headers, allowed: set[int], default: int,
                            compression: int = None) -> int:
        try:
            if compression is None:
                compression = await cls.propose_compression(src, headers, default)
            if compression not in allowed:
                raise ClientSupportError(f'Compression unsupported by client: {cls.compressors[compression].type_name}')

            await cls.compressors[compression].compress_file(src, dst, headers)
            return compression
        except (KeyError, ValueError, TypeError) as err:
            raise CompressorError(f'Failed to compress file: {str(err)}', data=src, headers=headers)

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, headers: T_Headers, compression: int) -> None:
        try:
            await cls.compressors[compression].decompress_file(src, dst, headers)
        except (KeyError, ValueError, TypeError) as err:
            raise CompressorError(f'Failed to decompress file: {str(err)}', data=src, headers=headers)

    @classmethod
    async def propose_compression(cls, buff: bytes | Path, headers: T_Headers, default: int):
        if isinstance(buff, bytes):
            ln = len(buff)
        elif isinstance(buff, Path):
            ln = buff.stat().st_size
        else:
            raise InvalidCompressorError('Unsupported buffer type', data=buff, headers=headers)
        if ln <= 4096:
            return C_NONE
        else:
            return default



import asyncio
import inspect
from collections import defaultdict
from typing import Any, Type

from cats.errors import CatsError
from cats.identity import IdentityObject
from cats.utils import require

__all__ = [
    'AuthError',
    'AuthMethod',
    'Auth',
]


class AuthError(CatsError, RuntimeError):
    pass


class AuthMethod:
    @classmethod
    async def sign_in(cls, **kwargs) -> tuple[IdentityObject, Any, int | float | None]:
        """
        Returns tuple[identity, credentials]. Consider to use returned credentials, instead of passed in
        :raises AuthError:
        """
        raise NotImplementedError


class Auth:
    def __init__(self, methods: list[Type[AuthMethod] | str]):
        self.methods: dict[frozenset, list[Type[AuthMethod]]] = defaultdict(list)
        for method_path in methods:
            if isinstance(method_path, str):
                method = require(method_path, strict=True)
            else:
                method = method_path
            assert isinstance(method, type) and issubclass(method, AuthMethod)
            params = frozenset(inspect.signature(method.sign_in).parameters.keys())
            self.methods[frozenset(params)].append(method)

    async def sign_in(self, **kwargs) -> tuple[IdentityObject, Any, int | float | None]:
        prev_err = None
        for method in self.methods.get(frozenset(kwargs.keys()), []):
            try:
                return await method.sign_in(**kwargs)
            except (KeyboardInterrupt, asyncio.CancelledError, asyncio.TimeoutError):
                raise
            except Exception as exc:
                try:
                    raise exc from prev_err
                except BaseException as err:
                    prev_err = err
        raise AuthError('Unable to authenticate') from prev_err

    async def sign_in_silent(self, **kwargs) -> tuple[IdentityObject | None, Any | None, int | float | None]:
        try:
            return await self.sign_in(**kwargs)
        except AuthError:
            return None, None, None



from pathlib import Path
from types import GeneratorType
from typing import AsyncIterable, Iterable, TypeAlias

import ujson

from cats.errors import MalformedHeadersError

try:
    from django.db.models import QuerySet, Model
except ImportError:
    QuerySet = type('QuerySet', (list,), {})
    Model = type('Model', (list,), {})

__all__ = [
    'Bytes',
    'BytesGen',
    'BytesAsyncGen',
    'BytesAnyGen',
    'Byte',
    'Json',
    'File',
    'List',
    'Missing',
    'MISSING',
    'QuerySet',
    'Model',
    'T_Headers',
    'Headers',
]

Bytes: TypeAlias = bytes | bytearray | memoryview
BytesGen: TypeAlias = Iterable[Bytes]
BytesAsyncGen: TypeAlias = AsyncIterable[Bytes]
BytesAnyGen: TypeAlias = BytesGen | BytesAsyncGen

Byte: TypeAlias = Bytes
Json: TypeAlias = str | int | float | dict | list | bool | None
File: TypeAlias = Path | str
List = list | tuple | set | GeneratorType | QuerySet


class Missing(str):
    """
    Custom Missing type is required for Pydantic to work properly. IDK
    """
    __slots__ = ()

    def __init__(self):
        super().__init__()

    def __eq__(self, other):
        return isinstance(other, Missing)

    def __bool__(self):
        return False


MISSING = Missing()


class Headers(dict):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        v = self._convert(*args, **kwargs)
        if (offset := v.get('offset', None)) and (not isinstance(offset, int) or offset < 0):
            raise MalformedHeadersError('Invalid offset header', headers=v)
        super().__init__(v)

    @classmethod
    def _key(cls, key: str) -> str:
        return key.replace(' ', '-').title()

    def __getitem__(self, item):
        return super().__getitem__(self._key(item))

    def __setitem__(self, key, value):
        return super().__setitem__(self._key(key), value)

    def __delitem__(self, key):
        return super().__delitem__(self._key(key))

    def __contains__(self, item):
        return super().__contains__(self._key(item))

    @classmethod
    def _convert(cls, *args, **kwargs):
        return {cls._key(k): v for k, v in dict(*args, **kwargs).items() if isinstance(k, str)}

    def update(self, *args, **kwargs) -> None:
        super().update(self._convert(*args, **kwargs))

    def encode(self) -> bytes:
        return ujson.dumps(self, ensure_ascii=False, escape_forward_slashes=False).encode('utf-8')

    @classmethod
    def decode(cls, headers: Bytes) -> 'Headers':
        try:
            headers = ujson.loads(headers)
        except ValueError:  # + UnicodeDecodeError
            headers = None
        return cls(headers or {})


T_Headers: TypeAlias = Headers | dict[str]



import asyncio
from dataclasses import dataclass
from typing import Type

from tornado.iostream import StreamClosedError

from cats.errors import CatsError
from cats.v2.handshake import Handshake

__all__ = [
    'Config',
]


@dataclass
class Config:
    idle_timeout: float | int = 120.0
    input_timeout: float | int = 120.0
    input_limit: int = 5
    debug: bool = False
    max_plain_payload: int = 16 * 1024 * 1024
    stream_errors: Type[Exception] | tuple[Type[Exception]] = (
        asyncio.TimeoutError,
        asyncio.CancelledError,
        asyncio.InvalidStateError,
        StreamClosedError,
    )
    ignore_errors: Type[Exception] | tuple[Type[Exception]] = (
        *stream_errors,
        CatsError,
        KeyboardInterrupt,
    )
    handshake: Handshake | None = None



import traceback
from logging import getLogger
from typing import Protocol

from richerr import RichErr
from tornado.iostream import StreamClosedError

from cats.v2.action import Action, ActionLike
from cats.v2.server.handlers import Handler

__all__ = [
    'Forward',
    'Middleware',
    'default_error_handler',
]

logging = getLogger('CATS')


class Forward(Protocol):
    async def __call__(self, handler: Handler) -> ActionLike | None: ...


class Middleware(Protocol):
    async def __call__(self, handler: Handler, forward: Forward) -> ActionLike | None: ...


async def default_error_handler(handler: Handler, forward: Forward) -> ActionLike | None:
    try:
        return await forward(handler)
    except (KeyboardInterrupt, StreamClosedError):
        raise
    except Exception as err:
        logging.debug(traceback.format_exc())
        err = RichErr.convert(err)
        return Action(err.dict(), status=err.code)



import asyncio
from logging import getLogger
from pathlib import Path
from time import time_ns
from typing import AsyncIterator, Type, TypeAlias, TypeVar

import struct_model

from cats.types import BytesAnyGen, BytesAsyncGen, Headers, T_Headers
from cats.v2.connection import Connection

__all__ = [
    'ActionLike',
    'Input',
    'BaseAction',
    'BasicAction',
    'Action',
    'StreamAction',
    'InputAction',
    'DownloadSpeedAction',
    'CancelInputAction',
    'PingAction',
]

logging = getLogger('CATS.conn')

MAX_IN_MEMORY: int
MAX_CHUNK_READ: int

ActionLike: TypeAlias = TypeVar('ActionLike', bound='Action')


class Input:
    __slots__ = ('future', 'conn', 'message_id', 'bypass_count', 'timer')

    def __init__(self,
                 future: asyncio.Future,
                 conn: Connection,
                 message_id: int,
                 bypass_count: bool = False,
                 timeout: float | int = None):
        self.timer: asyncio.Task | None
        self.future: asyncio.Future = future
        self.conn: Connection = conn
        self.message_id: int = message_id
        self.bypass_count: bool = bypass_count
        if timeout:
            self.timer = asyncio.get_running_loop().call_later(timeout, self.cancel)

    def done(self, result) -> None: ...

    def cancel(self) -> None: ...


class BaseAction(dict):
    __slots__ = ('data', 'headers', 'message_id', 'conn')
    __registry__: dict[bytes, Type['BaseAction']] = {}
    HEADER_SEPARATOR = b'\x00\x00'

    type_id: bytes
    Head: type
    status: property | int
    offset: property | int

    def __init__(self, data=None, *, headers: T_Headers = None, status: int = None,
                 message_id: int = None):
        self.data = data
        self.headers: Headers = Headers(headers or {})
        self.status: int = self.headers.get('Status', status or 200)
        self.message_id: int | None = message_id
        self.conn: Connection | None = None
        super().__init__()

    def __init_subclass__(cls, *, abstract: bool = False) -> None: ...

    @classmethod
    def get_class_by_type_id(cls, type_id: bytes) -> Type['BaseAction'] | None: ...

    async def ask(self, data=None, data_type: int = None, compression: int = None, *,
                  headers: T_Headers = None, status: int = None,
                  bypass_limit: bool = False, bypass_count: bool = False,
                  timeout: int | float = None) -> 'InputAction': ...

    @classmethod
    async def init(cls, conn: Connection) -> 'BaseAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'BaseAction.Head': ...

    async def dump_data(self, size: int) -> None: ...

    async def send(self, conn: Connection) -> None: ...

    def __repr__(self) -> str: ...


class BasicAction(BaseAction):
    __slots__ = ('data_len', 'data_type', 'compression', 'encoded')

    def __init__(self, data=None, *, headers: T_Headers = None, status: int = None,
                 message_id: int = None, data_len: int = None, data_type: int = None, compression: int = None,
                 encoded: int = None):
        self.data_len: int | None = data_len
        self.data_type: int | None = data_type
        self.compression: int | None = compression
        self.encoded: int | None = encoded
        super().__init__(data, headers=headers, status=status, message_id=message_id)

    async def recv_data(self) -> None: ...

    async def _recv_small_data(self) -> None: ...

    async def _recv_large_data(self) -> None: ...

    async def _recv_chunk(self, left: int) -> (bytes, int): ...

    async def _encode(self) -> tuple[Path | bytes, int, int, int]: ...

    async def _write_data_to_stream(self, conn: Connection, data: Path | bytes,
                                    data_len: int, data_type: int, compression: int) -> None: ...

    def __repr__(self) -> str: ...


class Action(BasicAction):
    class Head(struct_model.StructModel):
        handler_id: struct_model.uInt2
        message_id: struct_model.uInt2
        send_time: struct_model.uInt8
        data_type: struct_model.uInt1
        compression: struct_model.uInt1
        data_len: struct_model.uInt4

    __slots__ = ('handler_id', 'send_time')

    def __init__(self, data=None, *, headers: T_Headers = None, status: int = None, message_id: int = None,
                 handler_id: int = None, data_len: int = None, data_type: int = None, compression: int = None,
                 send_time: float = None, encoded: int = None):
        self.handler_id: int | None = handler_id
        self.send_time = send_time or (time_ns() // 1000000)
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         data_len=data_len, data_type=data_type, compression=compression, encoded=encoded)

    @classmethod
    async def init(cls, conn: Connection) -> 'Action': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'Action.Head': ...

    async def send(self, conn: Connection) -> None: ...

    def __repr__(self) -> str: ...


class StreamAction(Action):
    class Head(struct_model.StructModel):
        handler_id: struct_model.uInt2
        message_id: struct_model.uInt2
        send_time: struct_model.uInt8
        data_type: struct_model.uInt1
        compression: struct_model.uInt1

    def __init__(self, data=None, *, headers: T_Headers = None, status: int = None, message_id: int = None,
                 handler_id: int = None, data_type: int = None, compression: int = None,
                 send_time: float = None):
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         handler_id=handler_id, data_type=data_type, compression=compression,
                         send_time=send_time)

    @classmethod
    async def init(cls, conn: Connection) -> 'StreamAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'StreamAction.Head': ...

    async def recv_data(self) -> None: ...

    async def _recv_large_chunk(self, fh, chunk_size) -> int: ...

    async def _recv_small_chunk(self, fh, chunk_size) -> int: ...

    async def _encode_gen(self, conn: Connection) -> tuple[BytesAsyncGen, int]: ...

    async def send(self, conn: Connection) -> None: ...

    async def _write_data_to_stream(self, conn: Connection, data: BytesAsyncGen, *_, compression: int) -> None: ...

    async def _async_gen(self, gen: BytesAnyGen, chunk_size: int) -> AsyncIterator[bytes]: ...

    def __repr__(self) -> str: ...


class InputAction(BasicAction):
    class Head(struct_model.StructModel):
        message_id: struct_model.uInt2
        data_type: struct_model.uInt1
        compression: struct_model.uInt1
        data_len: struct_model.uInt4

    @classmethod
    async def init(cls, conn: Connection) -> 'InputAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'InputAction.Head': ...

    async def send(self, conn: Connection) -> None: ...

    async def reply(self, data=None, data_type: int = None, compression: int = None, *,
                    headers: T_Headers = None, status: int = None) -> ActionLike | None: ...

    async def cancel(self) -> ActionLike | None: ...


class DownloadSpeedAction(BaseAction):
    class Head(struct_model.StructModel):
        speed: struct_model.uInt4

    def __init__(self, speed: int):
        super().__init__()
        self.speed: int = speed

    @classmethod
    async def init(cls, conn: Connection) -> 'DownloadSpeedAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'DownloadSpeedAction.Head': ...

    async def send(self, conn: Connection) -> None: ...

    def __repr__(self) -> str: ...


class CancelInputAction(BaseAction):
    class Head(struct_model.StructModel):
        message_id: struct_model.uInt2

    @classmethod
    async def init(cls, conn: Connection) -> 'CancelInputAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'CancelInputAction.Head': ...

    async def send(self, conn: Connection) -> None: ...


class PingAction(BaseAction):
    class Head(struct_model.StructModel):
        send_time: struct_model.uInt8

    def __init__(self, send_time: int = None):
        super().__init__()
        self.recv_time: int = time_ns() // 1000000
        self.send_time: int = send_time or self.recv_time

    @classmethod
    async def init(cls, conn: Connection) -> 'PingAction': ...

    @classmethod
    async def _recv_head(cls, conn: Connection) -> 'PingAction.Head': ...

    async def send(self, conn: Connection) -> None: ...

    def __repr__(self) -> str: ...



class CatsError(Exception):
    """Base class for CATS errors"""
    __slots__ = ()


class InputCancelled(CatsError):
    """Raise if received CANCEL action"""
    __slots__ = ()


class ProtocolError(CatsError, ValueError, OSError):
    """Raise if protocol was violated"""
    __slots__ = ('conn',)

    def __init__(self, *args: object, conn) -> None:
        self.conn = conn
        super().__init__(*args)


class ClientSupportError(CatsError, RuntimeError):
    """Raised if server issued action, that client does not support"""
    __slots__ = ()


class ActionError(CatsError, RuntimeError):
    """Raised if client`s request violated handler rules"""
    __slots__ = ('action',)

    def __init__(self, *args: object, action) -> None:
        self.action = action
        super().__init__(*args)


class CatsUsageError(CatsError, RuntimeError):
    """Raise if cats API was violated"""
    __slots__ = ()


class UnsupportedSchemeError(CatsError, TypeError):
    """Raise if scheme is unsupported"""
    __slots__ = ('scheme',)

    def __init__(self, *args: object, scheme) -> None:
        self.scheme = scheme
        super().__init__(*args)


class ValidationError(CatsError, ValueError):
    """Raise if data does not fulfil requirements"""
    __slots__ = ('data',)

    def __init__(self, *args: object, data) -> None:
        self.data = data
        super().__init__(*args)


class CodecError(CatsError, ValueError):
    """Raise if codec failed to handle data"""
    __slots__ = ('data', 'headers')

    def __init__(self, *args: object, data, headers: dict) -> None:
        self.data = data
        self.headers = headers
        super().__init__(*args)


class InvalidCodecError(CatsError, TypeError):
    """Raise if codec failed to handle data"""
    __slots__ = ('data', 'headers')

    def __init__(self, *args: object, data, headers: dict) -> None:
        self.data = data
        self.headers = headers
        super().__init__(*args)


class CompressorError(CatsError, ValueError):
    """Raise if compressor failed to handle data"""
    __slots__ = ('data', 'headers')

    def __init__(self, *args: object, data, headers: dict) -> None:
        self.data = data
        self.headers = headers
        super().__init__(*args)


class InvalidCompressorError(CatsError, TypeError):
    """Raise if compressor failed to handle data"""
    __slots__ = ('data', 'headers')

    def __init__(self, *args: object, data, headers: dict) -> None:
        self.data = data
        self.headers = headers
        super().__init__(*args)


class MalformedDataError(CatsError, ValueError):
    """Raise if data is corrupted, but it is not Codec error"""
    __slots__ = ('data', 'headers')

    def __init__(self, *args: object, data, headers: dict) -> None:
        self.data = data
        self.headers = headers
        super().__init__(*args)


class MalformedHeadersError(CatsError, ValueError):
    """Raise if headers are corrupted"""
    __slots__ = ('headers',)

    def __init__(self, *args: object, headers: dict) -> None:
        self.headers = headers
        super().__init__(*args)


class HandshakeError(CatsError, ValueError):
    """Raise if handshake failed"""
    __slots__ = ('conn', 'handshake')

    def __init__(self, *args: object, conn, handshake) -> None:
        self.conn = conn
        self.handshake = handshake
        super().__init__(*args)



"""
proxy_protocol.py

Implements the HAProxy Proxy Protocol that passes the actual client
IP behind a load balancer to the server.

Note that the spec is defined here:
https://www.haproxy.org/download/1.8/doc/proxy-protocol.txt
"""
from functools import wraps
from typing import TypedDict

from tornado import httpserver, httputil, iostream, tcpserver
from tornado.iostream import IOStream

# Set the maximum number of bytes that should be allowed in the PROXY
# protocol; this is specifically designed so the proxy line fits in a
# single TCP packet, if I remember correctly.
MAX_PROXY_PROTO_BYTES = 108


class ProxyDict(TypedDict):
    client_ip: str
    client_port: int
    local_ip: str
    local_port: int


class ProxyProtocolInvalidAddress(Exception):
    """ Indicates that the proxy type was invalid.

    The proxy protocol requires fields like: TCP4 or TCP6 (for version 1)
    """
    pass


class ProxyProtocolInvalidLine(Exception):
    """ Indicates that the proxy line as a whole was invalid. """
    pass


class ProxyProtocolIncompleteLine(Exception):
    """ Indicates that the proxy line is incomplete. """
    pass


def parse_proxy_line(line: bytes | bytearray) -> tuple[ProxyDict, int]:
    """
    Parses the given line (string or sequence of bytes) for the client IP
    and other fields passed through the proxy protocol.

    This returns a tuple with elements as follows:
    (1) Dictionary with the parsed IP addresses
    (2) Index where the proxy protocol stopped parsing.

    This throws exceptions if the proxy protocol could not be parsed.
    """
    line_len = len(line)
    if not line.startswith(b"PROXY "):
        for i, ch in enumerate("PROXY "):
            if i >= line_len:
                raise ProxyProtocolIncompleteLine()
            if line[i] != ch:
                break
        raise ProxyProtocolInvalidLine()

    idx = line.find(b'\r\n')
    if idx < 0:
        if line_len >= MAX_PROXY_PROTO_BYTES:
            raise ProxyProtocolInvalidLine()
        raise ProxyProtocolIncompleteLine()

    # Now, try and parse the proxy line as desired.
    proxy_line = line[:idx]
    fields: list[bytes] = proxy_line.split(b' ')
    if len(fields) != 6:
        raise ProxyProtocolInvalidLine()

    # Check if the proxy line is for v1 of the protocol.
    if fields[0] == b"PROXY":
        try:
            if fields[1] == b"TCP4" or fields[1] == b"TCP6":
                source_addr = fields[2]
                dest_addr = fields[3]
                source_port = int(fields[4])
                dest_port = int(fields[5])
            else:
                raise ProxyProtocolInvalidAddress()
        except:
            raise ProxyProtocolInvalidAddress()

        return {
                   'client_ip': source_addr.decode('utf-8'),
                   'client_port': source_port,
                   'local_ip': dest_addr.decode('utf-8'),
                   'local_port': dest_port
               }, int(idx + 2)
    raise ProxyProtocolInvalidLine()


def putback_line_in_stream(line: bytes | bytearray, stream: IOStream):
    """
    HACKY: This call reinjects the given line back into the stream.

    This emulates a "putback" or "unget" operation. This is subject to
    break, though, so it should be used with caution, since it uses
    tornado internals.
    """
    # line_len = len(line)
    # stream._read_buffer_pos -= line_len
    # stream._read_buffer_size += line_len
    if not line:
        return
    stream._read_buffer[:0] = line
    stream._read_buffer_size += len(line)


async def parse_proxy_line_from_buff(stream: IOStream) -> ProxyDict:
    line = b''
    bytes_remaining = MAX_PROXY_PROTO_BYTES
    addr = None
    idx = 0
    # This "while" loop is necessary because Tornado does not (but it should!)
    # support an option to `stream.read_until()` to not automatically close the
    # connection if the maximum byte count is reached before finding the
    # delimiter.
    #
    # Again, @bdarnell, can you provide a cleaner way to do this?
    #
    while True:
        # Keep reading until we get the maximum number of bytes allowed.
        # We will break out of this loop early if we identify that the
        # parsed line doesn't imply the proxy protocol.
        val = await stream.read_bytes(bytes_remaining, partial=True)
        line += val
        bytes_remaining = MAX_PROXY_PROTO_BYTES - len(line)

        try:
            addr, idx = parse_proxy_line(line)
            break
        except ProxyProtocolIncompleteLine:
            # If the line is incomplete, keep trying to read.
            continue
        except (ProxyProtocolInvalidLine, ProxyProtocolInvalidAddress):
            idx = 0
            break

    # Reinsert any unparsed bytes back into the stream.
    putback_line_in_stream(line[idx:], stream)

    # Return the address we parsed.
    # This supports older tornado versions, though this can be very
    # easily fixed to work with asyncio.
    return addr


def handle_with_proxy(fn: tcpserver.TCPServer.handle_stream = None):
    if fn is None:
        return handle_with_proxy

    @wraps(fn)
    async def handle_stream(self, stream: IOStream, address: tuple[str, int]):
        try:
            addr = await parse_proxy_line_from_buff(stream)
            if addr is not None:
                address = (addr['client_ip'], addr['client_port'])
                stream.proxy_addr = addr
            return await fn(self, stream, address)
        except iostream.StreamClosedError:
            raise
        except:
            return await fn(self, stream, address)

    return handle_stream


class ProxyProtocolTCPServer(tcpserver.TCPServer):
    """
    Wrapper for tornado.tcpserver.TCPServer that parses out the
    HAProxy PROXY protocol and attaches the actual IP to the stream
    as an attribute named 'proxy_addr'.

    The proxy protocol spec is defined here:
    http://www.haproxy.org/download/1.8/doc/proxy-protocol.txt
    """

    async def handle_stream(self, stream: IOStream, address: tuple[str, int]):
        try:
            addr = await parse_proxy_line_from_buff(stream)
            if addr is not None:
                stream.proxy_addr = addr
        except iostream.StreamClosedError:
            raise
        finally:
            await super(ProxyProtocolTCPServer, self).handle_stream(stream, address)


class ProxyProtocolAdapter(httputil.HTTPMessageDelegate, object):
    """
    Implements a HTTPMessageDelegate that injects the real client IP into the
    request context for each request.

    Basically, this call makes sure that the following code actually returns
    the correct client IP:
    `
    class SampleHandler(tornado.web.RequestHandler):
        def post(self):
            self.request.remote_ip
    `

    This works by intercepting the delegate that is called whenever a request
    is made, then injecting the proper remote_ip (and other relevant fields).
    When the request is finished, or stuff falls out of scope, this "undoes"
    the injected changes so everything is consistent.

    Note that this particular adapter only works for HTTP (and by extension
    WebSocket) calls.
    """

    def __init__(self, delegate, request_conn, proxy_ip=None):
        self.connection = request_conn
        self.delegate = delegate
        self.proxy_ip = proxy_ip

    def _apply_proxy_info(self):
        self._orig_ip = self.connection.context.remote_ip
        if self.proxy_ip:
            self.connection.context.remote_ip = self.proxy_ip

    def _undo_proxy_info(self):
        self.connection.context.remote_ip = self._orig_ip

    def headers_received(self, start_line, headers):
        self._apply_proxy_info()
        return self.delegate.headers_received(start_line, headers)

    def data_received(self, chunk):
        return self.delegate.data_received(chunk)

    def finish(self):
        self.delegate.finish()
        self._undo_proxy_info()

    def on_connection_close(self):
        self.delegate.on_connection_close()
        self._undo_proxy_info()


class ProxyProtocolHTTPServer(httpserver.HTTPServer):
    """
    Wrapper for tornado.httpserver.HTTPServer that parses out the
    HAProxy PROXY protocol and sets the appropriate remote_ip as
    defined here:
    http://www.haproxy.org/download/1.8/doc/proxy-protocol.txt
    """

    async def handle_stream(self, stream: IOStream, address: tuple[str, int]):
        """
        Creates the stream for this connection.

        This parses out the proxy protocol from the request and injects the
        relevant fields onto the given stream object at: `stream.proxy_addr`
        """
        try:
            addr = await parse_proxy_line_from_buff(stream)
            if addr is not None:
                stream.proxy_addr = addr

            # Use our parsed address here.
            super(ProxyProtocolHTTPServer, self).handle_stream(stream, address)
        except iostream.StreamClosedError:
            pass
        except:
            super(ProxyProtocolHTTPServer, self).handle_stream(stream, address)

    def start_request(self, server_conn, request_conn):
        """
        Override so that we can inject the real ip for `self.request.remote_ip`
        in each request.
        """
        delegate = super(ProxyProtocolHTTPServer, self).start_request(server_conn, request_conn)

        # If the real client IP was parsed via the proxy protocol, the address would
        # have been attached to the socket in the 'proxy_addr' attribute, so we try
        # to fetch it. If we couldn't, just ignore it and default to the usual
        # behavior. (Should we log it instead?)
        try:
            proxy_addr = server_conn.stream.proxy_addr
            addr = proxy_addr['client_ip']
            return ProxyProtocolAdapter(delegate, request_conn, proxy_ip=addr)
        except Exception as e:
            return delegate



import asyncio
import inspect

from cats.v2 import Config
from cats.v2.action import Action
from cats.v2.client import Connection as BaseConn

__all__ = [
    'Connection',
]


class Connection(BaseConn):
    __slots__ = ('broadcast_inbox', 'store_handled_broadcast', 'broadcast_inbox_size')

    def __init__(self, conf: Config, api_version: int):
        super().__init__(conf, api_version)
        self.broadcast_inbox: list[Action] = []
        self.store_handled_broadcast: bool = False
        self.broadcast_inbox_size: int = 5

    async def handle_broadcast(self, action: Action):
        if action.handler_id in self.subscriptions:
            for fn in self.subscriptions[action.handler_id].values():
                res = fn(action)
                if inspect.isawaitable(res):
                    await res
            if not self.store_handled_broadcast:
                return
        self.broadcast_inbox.append(action)
        self.broadcast_inbox = self.broadcast_inbox[-self.broadcast_inbox_size:]

    def clear_broadcast_inbox(self):
        self.broadcast_inbox.clear()

    def get_received_broadcast(self, handler_id: int) -> (int | None, Action | None):
        for i, action in enumerate(self.broadcast_inbox):
            if action.handler_id == handler_id:
                return i, action
        return None, None

    async def recv_broadcast(self, handler_id: int) -> Action | None:
        i, action = self.get_received_broadcast(handler_id)
        if action is not None:
            if not self.store_handled_broadcast:
                self.broadcast_inbox.pop(i)
            return action
        fut = asyncio.Future()
        sub_id = self.subscribe(handler_id, lambda res: fut.set_result(res))
        response = await asyncio.wait_for(fut, 5.0)
        self.unsubscribe(handler_id, sub_id)
        return response



import asyncio
import inspect
from collections import defaultdict
from logging import getLogger
from random import randint
from time import time, time_ns
from typing import Awaitable, Callable

from tornado.iostream import IOStream
from tornado.tcpclient import TCPClient

from cats.errors import ProtocolError
from cats.types import BytesAnyGen
from cats.utils import as_uint, to_uint
from cats.v2.action import Action, ActionLike, BaseAction, PingAction, StreamAction
from cats.v2.config import Config
from cats.v2.connection import Connection as BaseConnection
from cats.v2.statement import ClientStatement, ServerStatement

__all__ = [
    'Connection',
]


class Connection(BaseConnection):
    __slots__ = (
        'api_version',
        'time_delta',
        'subscriptions',
        'address',
        '_sub_id',
        '_stream',
        '_listener',
        '_sender',
        '_pinger',
        '_recv_pool',
    )
    PROTOCOL_VERSION: int = 2
    logging = getLogger('CATS.Client')

    def __init__(self, conf: Config, api_version: int):
        super().__init__(conf)
        self.api_version: int = api_version
        self.time_delta: float = 0.0
        self.subscriptions: dict[
            int, dict[int, Callable[[Action], Awaitable[None] | None]]] = defaultdict(
            dict
        )
        self._sub_id: int = 0
        self._listener: asyncio.Task | None = None
        self._sender: asyncio.Task | None = None
        self._pinger: asyncio.Task | None = None
        self._recv_pool: dict[int, asyncio.Future] = {}
        self._stream: IOStream | None = None
        self.address: tuple[str, int] = '0.0.0.0', 0

    async def connect(self, host: str, port: int, **kwargs):
        client = TCPClient()
        self._stream = await client.connect(host, port, **kwargs)
        self.address = host, port
        await self.init()
        self._listener = asyncio.get_running_loop().create_task(self.start())
        self._listener.add_done_callback(self.on_tick_done)
        self.debug(f'New connection established: {self.address}')

    async def start(self) -> None:
        self._sender = self._loop.create_task(self.send_loop())
        self._pinger = self._loop.create_task(self.ping())

        self._sender.add_done_callback(self.on_tick_done)
        self._pinger.add_done_callback(self.on_tick_done)

        await self.recv_loop()

    async def init(self):
        await self.write(to_uint(self.PROTOCOL_VERSION, 4))
        result = await self.read(4)
        if result != bytes(4):
            raise ProtocolError(
                f'Unsupported protocol version. '
                f'Please upgrade your client to: {as_uint(result)}',
                conn=self
            )

        client_stmt = ClientStatement(
            api=self.api_version,
            client_time=time_ns() // 1000_000,
            scheme_format='JSON',
            compressors=['gzip', 'zlib'],
            default_compression='zlib',
        )
        self.set_compressors(['gzip', 'zlib'], 'zlib')
        await self.write(client_stmt.pack())
        stmt: ServerStatement = ServerStatement.unpack(
            await self.read(as_uint(await self.read(4)))
        )
        self.time_delta = (stmt.server_time / 1000) - time()

        if self.conf.handshake is not None:
            await self.conf.handshake.send(self)

        self.debug(f'{self} initialized')

    async def handle(self, action: BaseAction):
        if isinstance(action, PingAction):
            await self.handle_ping_action(action)
        elif isinstance(action, Action):
            await action.recv_data()
            if action.message_id >= 0x8000:
                await self.handle_broadcast(action)
            elif action.message_id in self._recv_pool:
                await self.handle_response(action)
            else:
                self.debug(f'Received unexpected action {action = }')
        else:
            self.debug(f'Received unsupported Action: {type(action).__qualname__}')

    async def handle_response(self, action: Action):
        future = self._recv_pool.pop(action.message_id)
        future.set_result(action)

    async def handle_broadcast(self, action: Action):
        if action.handler_id in self.subscriptions:
            for fn in self.subscriptions[action.handler_id].values():
                res = fn(action)
                if inspect.isawaitable(res):
                    await res

    async def handle_ping_action(self, action: PingAction):
        self.debug(f'Pong {action.send_time} [-] {action.recv_time}')
        await action.dump_data(0)

    def subscribe(
        self,
        handler_id: int,
        handler: Callable[[Action], Awaitable[None] | None]
    ) -> int:
        self._sub_id += 1
        self.subscriptions[handler_id][self._sub_id] = handler
        return self._sub_id

    def unsubscribe(
        self,
        handler_id: int,
        handler: Callable[[Action], Awaitable[None] | None] | int
    ) -> None:
        if handler_id in self.subscriptions:
            if isinstance(handler, int):
                self.subscriptions[handler_id].pop(handler, None)
            else:
                self.subscriptions[handler_id] = {
                    k: v for k, v in self.subscriptions[handler_id]
                    if v is not handler
                }

    async def set_download_speed(self, speed=0):
        await self.write(b'\x05')
        await self.write(to_uint(speed, 4))

    async def send(
        self,
        handler_id: int,
        data=None,
        message_id=None,
        compression=None, *,
        headers=None, status=None
    ) -> ActionLike | None:
        action = Action(
            data=data, headers=headers, status=status,
            message_id=self.get_free_message_id() if message_id is None else message_id,
            handler_id=handler_id, compression=compression
        )
        await action.send(self)
        return await self.recv(action.message_id)

    async def send_stream(
        self,
        handler_id: int,
        data: BytesAnyGen,
        data_type: int,
        message_id=None,
        compression=None, *,
        headers=None,
        status=None
    ) -> ActionLike | None:
        action = StreamAction(
            data=data, headers=headers, status=status,
            message_id=self.get_free_message_id() if message_id is None else message_id,
            handler_id=handler_id, data_type=data_type, compression=compression
        )
        await action.send(self)
        return await self.recv(action.message_id)

    async def recv(self, message_id: int) -> Action | None:
        future = asyncio.Future()
        self._recv_pool[message_id] = future
        return await future

    async def ping(self):
        if not self.conf.idle_timeout:
            return
        wait = max(0.1, round(self.conf.idle_timeout * 0.9, 2))
        pong = PingAction()
        while self.is_open:
            await asyncio.sleep(wait)
            await pong.send(self)

    def __str__(self) -> str:
        return f'CATS.Connection: {self.host}:{self.port} api@{self.api_version}'

    def get_free_message_id(self) -> int:
        while True:
            message_id = randint(0x0000, 0x7FFF)
            if message_id not in self._message_pool:
                return message_id

    def _close_tasks(self):
        yield from super()._close_tasks()
        yield from self._recv_pool.values()
        yield self._pinger
        yield self._sender



import warnings
from typing import Type

from pytest import PytestWarning, fixture, mark

from cats.test_utils.client import Connection
from cats.v2 import Config, Handshake
from cats.v2.auth import Auth
from cats.v2.server import Api, Application, Connection as ServerConnection, Middleware, Server, default_error_handler

__all__ = [
    'cats_handshake',
    'cats_apis',
    'cats_auth',
    'cats_middleware',
    'cats_config',
    'cats_server_connection',
    'cats_api_version',
    'cats_app',
    'cats_server',
    'cats_conn',
]


@fixture(scope='session')
def cats_handshake() -> Handshake:
    raise NotImplementedError('Please, overwrite "cats_handshake" fixture in order for CATS tests to work')


@fixture(scope='session')
def cats_apis() -> list[Api]:
    raise NotImplementedError('Please, overwrite "cats_apis" fixture in order for CATS tests to work')


@fixture(scope='session')
def cats_auth() -> Auth | None:
    raise NotImplementedError('Please, overwrite "cats_auth" fixture in order for CATS tests to work')


@fixture(scope='session')
def cats_middleware() -> list[Middleware]:
    warnings.warn(PytestWarning('You may want to overwrite "cats_middleware" fixture'))
    return [
        default_error_handler,
    ]


@fixture(scope='function')
def cats_api_version() -> int:
    warnings.warn(PytestWarning('You may want to overwrite "cats_api_version" fixture'))
    return 1


@fixture(scope='session')
def cats_config(cats_handshake) -> Config:
    return Config(
        idle_timeout=10.0,
        input_timeout=10.0,
        debug=True,
        handshake=cats_handshake,
    )


@fixture(scope='session')
def cats_server_connection() -> Type[ServerConnection] | None:
    return None


@fixture(scope='session')
def cats_app(cats_apis, cats_middleware, cats_config, cats_auth, cats_server_connection):
    return Application(
        apis=cats_apis,
        middleware=cats_middleware,
        auth=cats_auth,
        config=cats_config,
        connection=cats_server_connection,
    )


@fixture(scope='session')
@mark.asyncio
async def cats_server(cats_app) -> Server:
    """
    Runs an TCP server for each module and return port
    :return:
    """

    server = Server(cats_app)
    server.bind_unused_port()
    server.start(1)
    yield server
    await server.shutdown()


@fixture
@mark.asyncio
async def cats_conn(cats_config, cats_api_version, cats_server) -> Connection:
    """
    Return TCP connection to test TCP server
    """
    conn = Connection(cats_config, cats_api_version)
    await conn.connect('127.0.0.1', cats_server.port, timeout=5)
    async with conn:
        yield conn



import asyncio
import socket
import ssl
from contextlib import asynccontextmanager
from logging import getLogger
from typing import Callable

from tornado.iostream import IOStream
from tornado.tcpserver import TCPServer
from tornado.testing import bind_unused_port

from cats.errors import CatsError
from cats.utils import as_uint, to_uint
from cats.v2.connection import ConnType, Connection
from cats.v2.server.application import Application
from cats.v2.server.connection import Connection as ServerConnection

from cats.v2.server.proxy import handle_with_proxy

__all__ = [
    'Server',
]

logging = getLogger('CATS.Server')


class Server(TCPServer):
    __slots__ = ('app', 'port', 'connections')
    protocols: tuple[int, int] = 2, 2
    instances: list['Server'] = []

    def __init__(
        self, app: Application,
        ssl_options: dict[str] | ssl.SSLContext | None = None,
        max_buffer_size: int | None = None,
        read_chunk_size: int | None = None
    ):
        self.app: Application = app
        self.port: int | None = None
        self.connections: list[Connection] = []
        super().__init__(
            ssl_options=ssl_options,
            max_buffer_size=max_buffer_size,
            read_chunk_size=read_chunk_size
        )

    @classmethod
    async def broadcast(
        cls,
        channel: str,
        handler_id: int,
        data=None,
        message_id: int = None,
        compression: int = None,
        *, headers=None, status: int = None
    ):
        return await asyncio.gather(
            *(
                conn.send(
                    handler_id,
                    data,
                    message_id,
                    compression,
                    headers=headers,
                    status=status
                )
                for server in cls.running_servers()
                for conn in server.app.channel(channel)
            )
        )

    @classmethod
    async def conditional_broadcast(
        cls,
        channel: str,
        _filter: Callable[['Server', Connection], bool],
        handler_id: int,
        data=None,
        message_id: int = None,
        compression: int = None,
        *, headers=None, status: int = None
    ):
        return await asyncio.gather(
            *(
                conn.send(
                    handler_id,
                    data,
                    message_id,
                    compression,
                    headers=headers,
                    status=status
                )
                for server in cls.running_servers()
                for conn in server.app.channel(channel)
                if _filter(server, conn)
            )
        )

    @handle_with_proxy
    async def handle_stream(self, stream: IOStream, address: tuple[str, int]) -> None:
        try:
            protocol_version = as_uint(await stream.read_bytes(4))
            if not self.protocols[0] <= protocol_version <= self.protocols[1]:
                await stream.write(to_uint(self.protocols[1], 4))
                stream.close(CatsError('Unsupported protocol version'))
                return
            await stream.write(bytes(4))
            async with self.create_connection(stream, address, protocol_version) as conn:
                conn: ServerConnection
                conn.debug(f'[INIT {address}]')
                await conn.init()
                await conn.start()
            conn.debug(f'[STOP {address}]')
        except self.app.config.stream_errors:
            pass

    @asynccontextmanager
    async def create_connection(
        self,
        stream: IOStream,
        address: tuple[str, int],
        protocol: int,
    ) -> ConnType:
        conn_class = self.app.ConnectionClass or ServerConnection
        conn = conn_class(stream, address, protocol, self.app.config, self.app)
        try:
            self.connections.append(conn)
            self.app.attach_conn_to_channel(conn, '__all__')
            async with conn:
                yield conn
        except (KeyboardInterrupt, asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except self.app.config.ignore_errors:
            pass
        finally:
            self.app.remove_conn_from_channels(conn)
            try:
                self.connections.remove(conn)
            except ValueError:
                pass

    @classmethod
    def running_servers(cls) -> list['Server']:
        return [server for server in cls.instances if server.is_running]

    @property
    def is_running(self) -> bool:
        return self._started and not self._stopped

    async def shutdown(self, exc=None):
        for conn in self.connections:
            conn.close(exc=exc)

        self.app.clear_all_channels()
        self.connections.clear()
        logging.info('Shutting down TCP Server')
        self.stop()

    def start(self, num_processes: int = 1, max_restarts: int = None) -> None:
        super().start(num_processes, max_restarts)

    def bind_unused_port(self):
        sock, port = bind_unused_port()
        self.add_socket(sock)
        self.port = port
        logging.info(f'Starting server at 127.0.0.1:{port}')

    def bind(
        self, port: int, address: str = None,
        family: socket.AddressFamily = socket.AF_UNSPEC,
        backlog: int = 128, reuse_port: bool = False
    ) -> None:
        super().bind(port, address, family, backlog, reuse_port)
        self.port = port
        logging.info(f'Starting server at {address}:{port}')

    def listen(self, port: int, address: str = "") -> None:
        super().listen(port, address)
        self.port = port
        logging.info(f'Starting server at {address}:{port}')

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        cls.instances.append(obj)
        return obj

    def __del__(self):
        self.instances.remove(self)



import asyncio
import functools
from logging import getLogger
from random import randint
from time import time_ns
from typing import Iterable

import rollbar
from tornado.iostream import IOStream

from cats.errors import ProtocolError
from cats.identity import Identity, IdentityObject
from cats.types import BytesAnyGen
from cats.utils import as_uint
from cats.v2.action import *
from cats.v2.auth import AuthError
from cats.v2.config import Config
from cats.v2.connection import Connection as BaseConnection
from cats.v2.server.application import Application
from cats.v2.server.handlers import HandlerItem
from cats.v2.statement import ClientStatement, ServerStatement

__all__ = [
    'Connection',
]


class Connection(BaseConnection):
    __slots__ = (
        'api_version',
        'client',
        'address',
        'protocol',
        '_stream',
        '_app',
        '_idle_timer',
    )

    logging = getLogger('CATS.Connection')

    def __init__(self, stream: IOStream, address: tuple[str, int], protocol: int, conf: Config, app: Application):
        super().__init__(conf)
        self.api_version: int | None = None
        self.address: tuple[str, int] = address
        self.protocol: int = protocol
        self._stream: IOStream = stream
        self._app = app
        self._credentials = None
        self._idle_timer: asyncio.Future | None = None
        self.client: ClientStatement | None = None
        self.debug(f'New connection established: {address}')

    @property
    def app(self):
        return self._app

    async def init(self):
        self.client: ClientStatement = ClientStatement.unpack(
            await self.read(as_uint(await self.read(4)))
        )
        self.api_version = self.client.api
        self.set_compressors(self.client.compressors, self.client.default_compression)
        self.debug(f'[RECV {self.address}] {self.client}')

        server_stmt = ServerStatement(
            server_time=time_ns() // 1000_000,
        )
        await self.write(server_stmt.pack())
        self.debug(f'[SEND {self.address}] {server_stmt}')

        if self.conf.handshake is not None:
            await self.conf.handshake.validate(self)

        self.debug(f'{self} initialized')

    async def handle(self, action: BaseAction):
        if isinstance(action, InputAction):
            await action.recv_data()
            if action.message_id in self.input_pool:
                self.input_pool[action.message_id].done(action)
            else:
                raise ProtocolError('Received answer but input does`t exists', conn=self)
        elif isinstance(action, CancelInputAction):
            if action.message_id in self.input_pool:
                self.input_pool[action.message_id].cancel()
            await action.dump_data(0)
        elif isinstance(action, DownloadSpeedAction):
            limit = action.speed
            if not limit or (1024 <= limit <= 33_554_432):
                self.download_speed = limit
            else:
                raise ProtocolError('Unsupported download speed limit', conn=self)
            await action.dump_data(0)
        elif isinstance(action, PingAction):
            self.debug(f'Ping {action.send_time} [-] {action.recv_time}')
            await action.send(self)
            await action.dump_data(0)
        elif isinstance(action, Action):
            async with self.preserve_message_id(action.message_id):
                handler = self.dispatch(action.handler_id)
                try:
                    result = await self.app.run(handler(action))
                    if result is not None:
                        if not isinstance(result, Action):
                            raise ProtocolError('Returned invalid response', conn=self)

                        result.handler_id = action.handler_id
                        result.message_id = action.message_id
                        result.offset = action.offset
                        await result.send(self)
                except action.conn.conf.ignore_errors:
                    raise
                except Exception:
                    rollbar.report_exc_info(extra_data=self.identity_scope_user)
                    raise

    def dispatch(self, handler_id):
        handlers = self.app.get_handlers_by_id(handler_id)
        if isinstance(handlers, HandlerItem):
            return handlers.handler

        elif isinstance(handlers, list):
            for item in handlers:
                item: HandlerItem
                end_version = self.api_version if item.end_version is None else item.end_version
                if item.version <= self.api_version <= end_version:
                    return item.handler

        raise ProtocolError(f'Handler with id {handler_id} not found', conn=self)

    async def send(self, handler_id: int, data=None, message_id=None, compression=None, *,
                   headers=None, status=None):
        action = Action(data=data, headers=headers, status=status,
                        message_id=self.get_free_message_id() if message_id is None else message_id,
                        handler_id=handler_id, compression=compression)
        await action.send(self)

    async def send_stream(self, handler_id: int, data: BytesAnyGen, data_type: int,
                          message_id=None, compression=None, *,
                          headers=None, status=None):
        action = StreamAction(data=data, headers=headers, status=status,
                              message_id=self.get_free_message_id() if message_id is None else message_id,
                              handler_id=handler_id, data_type=data_type, compression=compression)
        await action.send(self)

    def attach_to_channel(self, channel: str):
        self.app.attach_conn_to_channel(self, channel=channel)

    def detach_from_channel(self, channel: str):
        self.app.detach_conn_from_channel(self, channel=channel)

    @property
    def conns_with_same_identity(self) -> Iterable['Connection']:
        return self.app.channel(f'model_{self._identity.model_name}:{self._identity.id}')

    @property
    def conns_with_same_model(self) -> Iterable['Connection']:
        return self.app.channel(f'model_{self._identity.model_name}')

    async def sign_in(self, silent: bool = False, **kwargs) -> IdentityObject | None:
        try:
            identity, credentials, timeout = await self._app.auth.sign_in(**kwargs)
            self.set_identity(identity, credentials=credentials, timeout=timeout)
            return identity
        except AuthError:
            if not silent:
                raise

    def set_identity(self, identity: Identity, credentials=None, timeout: int | float | None = None):
        super().set_identity(identity, credentials=credentials, timeout=timeout)

        model_group = f'model_{identity.model_name}'
        auth_group = f'{model_group}:{identity.id}'
        self.attach_to_channel(model_group)
        self.attach_to_channel(auth_group)

    def sign_out(self):
        if not self.signed_in:
            return
        model_group = f'model_{self.identity.model_name}'
        auth_group = f'{model_group}:{self._identity.id}'

        self.detach_from_channel(auth_group)
        self.detach_from_channel(model_group)
        super().sign_out()

    def __str__(self) -> str:
        return f'CATS.Connection: {self.host}:{self.port} api@{self.api_version}'

    async def read(self, num_bytes: int, partial: bool = False) -> bytes:
        res = await self._stream.read_bytes(num_bytes, partial=partial)
        self.reset_idle_timer()
        return res

    async def read_until(self, delimiter: bytes, max_bytes: int | None) -> bytes:
        res = await self._stream.read_until(delimiter, max_bytes=max_bytes)
        self.reset_idle_timer()
        return res

    async def write(self, data: bytes | bytearray | memoryview) -> None:
        res = await self._stream.write(data)
        self.reset_idle_timer()
        return res

    def reset_idle_timer(self):
        if not self.conf.idle_timeout > 0:
            return
        if self._idle_timer is not None:
            self._idle_timer.cancel()

        self._idle_timer = self._loop.call_later(
            self.conf.idle_timeout,
            functools.partial(self.close, asyncio.TimeoutError())
        )

    def get_free_message_id(self) -> int:
        while True:
            message_id = randint(0x8000, 0xFFFF)
            if message_id not in self._message_pool:
                return message_id

    def _close_tasks(self):
        yield self._idle_timer
        yield from super()._close_tasks()



import asyncio
from collections import defaultdict
from dataclasses import dataclass
from logging import getLogger
from random import random
from time import time
from typing import Awaitable, Type

from cats.errors import ActionError
from cats.identity import Identity, IdentityObject
from cats.plugins import Form, Scheme, SchemeTypes, scheme_json, scheme_load
from cats.types import Json, List, T_Headers
from cats.v2.action import Action, ActionLike, InputAction
from cats.v2.codecs import T_FILE, T_JSON

__all__ = [
    'HandlerItem',
    'Api',
    'Handler',
]

logging = getLogger('CATS')


@dataclass
class HandlerItem:
    id: int
    name: str
    handler: Type['Handler']
    version: int | None = None
    end_version: int | None = None


class Api:
    __slots__ = ('_handlers',)

    def __init__(self):
        self._handlers: dict[int, list[HandlerItem]] = defaultdict(list)

    def register(self, handler: HandlerItem):
        if handler in self._handlers[handler.id]:
            return

        assert handler.version is None or handler.end_version is None or handler.version <= handler.end_version, \
            f'Invalid version range for handler {handler.id}: [{handler.version}..{handler.end_version}]'

        if handler.version is not None or handler.end_version is not None:
            assert handler.version is not None, f'Initial version is not provided for {handler}'

            try:
                last_handler = self._handlers[handler.id][-1]
                assert last_handler.version is not None or last_handler.end_version is not None, \
                    f'Attempted to add versioned {handler} to wildcard'

                if last_handler.end_version is not None:
                    assert last_handler.end_version < handler.version, \
                        f'Handler {handler} overlap {last_handler} version'
                else:
                    assert last_handler.version < handler.version, \
                        f'Handler {handler} overlap {last_handler} version'
                    last_handler.end_version = handler.version - 1
            except IndexError:
                pass
        self._handlers[handler.id].append(handler)

    @property
    def handlers(self):
        return self._handlers

    def update(self, app: 'Api'):
        self._handlers.update(app.handlers)

    def compute(self) -> dict[int, HandlerItem | list[HandlerItem]]:
        result = {}
        for handler_id, handler_list in self._handlers.items():
            if not handler_list:
                continue

            elif len(handler_list) == 1 and handler_list[0].version is None and handler_list[0].end_version is None:
                result[handler_id] = handler_list[0]
            else:
                result[handler_id] = handler_list

        return result


class Handler:
    __slots__ = ('action',)
    handler_id: int

    Loader: Scheme | None = None
    Dumper: Scheme | None = None

    data_type: int | tuple[int] | None = None
    min_data_len: int | None = None
    max_data_len: int | None = None
    block_models: tuple[str] | None = None
    require_models: tuple[str] | None = None
    require_auth: bool | None = None
    min_file_size: int | None = None
    max_file_size: int | None = None
    min_file_total_size: int | None = None
    max_file_total_size: int | None = None
    min_file_amount: int | None = None
    max_file_amount: int | None = None
    fix_exec_time: int | float | None = None

    def __init__(self, action: Action):
        self.action = action

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls, /, api: Api = None, id: int = None,
                          name: str = None, version: int = None, end_version: int = None):
        # abstract
        if api is None:
            return

        assert id is not None

        assert cls.Loader is None or (isinstance(cls.Loader, type) and issubclass(cls.Loader, SchemeTypes)), \
            'Handler.Loader must be subclass of BaseSerializer | BaseModel'
        assert cls.Dumper is None or (isinstance(cls.Dumper, type) and issubclass(cls.Dumper, SchemeTypes)), \
            'Handler.Dumper must be subclass of BaseSerializer | BaseModel'

        assert not (cls.require_auth is False and cls.require_models is not None), \
            f'{cls!s}.require_auth is False and {cls!s}.require_models is not None'

        if cls.require_auth is None and cls.require_models is not None:
            cls.require_auth = True
        api.register(HandlerItem(id, name, cls, version, end_version))  # noqa, pycharm bug
        cls.handler_id = id

    @property
    def conn(self):
        return self.action.conn

    async def __call__(self) -> ActionLike | None:
        st = time()
        res = await self.prepare()
        if not isinstance(res, Action):
            res = await self.handle()
        if self.fix_exec_time is not None:
            sp = time() - st
            if sp < self.fix_exec_time:
                await asyncio.sleep(self.fix_exec_time - sp - random() / 20)
            elif sp > self.fix_exec_time + 0.2:
                logging.warning(f'{type(self).__qualname__} run time exceeded fixed: {sp:.3f}/{self.fix_exec_time:.3f}')
        return res

    async def prepare(self) -> None:
        """Called before handler() method"""
        try:
            self._check_before_recv()
        except self.action.conn.conf.stream_errors:
            raise
        except BaseException:
            await self.action.dump_data(self.action.data_len)
            raise
        else:
            await self.action.recv_data()
            self._check_after_recv()

    async def handle(self):
        raise NotImplementedError

    async def json_load(self, *, many: bool = False, plain: bool = False) -> Json | Form:
        if self.action.data_type != T_JSON:
            raise TypeError('Unsupported data type. Expected JSON')

        data = self.action.data
        if self.Loader is None:
            return data
        if many is None:
            many = isinstance(data, List)
        return scheme_load(self.Loader, data, many=many, plain=plain)

    async def json_dump(self, data, *, headers: T_Headers = None,
                        status: int = 200, many: bool = None, plain: bool = False) -> ActionLike:
        if self.Dumper is not None:
            if not plain:
                if many is None:
                    many = isinstance(data, List)
                data = scheme_json(self.Dumper, data, many=many, plain=False)
                return Action(data=data, headers=headers, status=status, encoded=T_JSON)
            elif not isinstance(data, self.Dumper):
                raise TypeError('Resulted plain data does not match Dumper')
        return Action(data=data, headers=headers, status=status)

    @property
    def identity(self) -> Identity | IdentityObject | None:
        return self.action.conn.identity

    def ask(self, data=None, data_type: int = None, compression: int = None, *,
            headers: T_Headers = None, status: int = 200,
            bypass_limit=False, bypass_count=False, timeout=None) -> Awaitable['InputAction']:
        return self.action.ask(data, data_type=data_type, compression=compression,
                               headers=headers, status=status,
                               bypass_limit=bypass_limit, bypass_count=bypass_count, timeout=timeout)

    def _check_before_recv(self):
        self._check_data_type()
        self._check_data_len()
        self._check_models()
        self._check_auth()
        self._check_file_size()
        self._check_file_total_size()
        self._check_file_amount()

    def _check_after_recv(self):
        self._check_data_len()

    def _check_data_type(self):
        if (types := self.data_type) is None:
            return
        if isinstance(types, int):
            types = (types,)
        if self.action.data_type not in types:
            raise ActionError('Received payload type is not acceptable', action=self.action)

    def _check_data_len(self):
        if (x := self.action.data_len) is None:
            return
        min_len = self.min_data_len
        max_len = self.max_data_len
        if min_len is not None and x < min_len:
            raise ActionError(f'Received payload data size is less than allowed [{min_len}]', action=self.action)
        if max_len is not None and max_len < x:
            raise ActionError(f'Received payload data size is more than allowed [{max_len}]', action=self.action)

    def _check_models(self):
        model = self.identity.model_name if self.identity else None
        if self.block_models is not None and model in self.block_models:
            raise ActionError(f'Model {model} is forbidden', action=self.action)
        if self.require_models is not None and model not in self.require_models:
            raise ActionError(f'Model of {self.require_models} is required, but {model} presented', action=self.action)

    def _check_auth(self):
        if self.require_auth is None:
            return
        if self.require_auth ^ self.action.conn.signed_in:
            reason = 'required' if self.require_auth else 'forbidden'
            raise ActionError(f'Authentication is {reason}', action=self.action)

    def _check_file_size(self):
        if self.action.data_type != T_FILE:
            return
        for i, file in enumerate(self.action.headers.get('Files', [])):
            x = int(file['size'])
            self._check_min_max(x, self.min_file_size, self.max_file_size, f'File[{i}].size')

    def _check_file_total_size(self):
        if self.action.data_type != T_FILE:
            return

        if self.action.data_len is not None:
            x = self.action.data_len
        else:
            x = sum(file['size'] for file in self.action.headers.get('Files', []))

        self._check_min_max(x, self.min_file_total_size, self.max_file_total_size, '∑ File[n].size')

    def _check_file_amount(self):
        if self.action.data_type != T_FILE:
            return

        x = len(self.action.headers.get('Files', []))
        self._check_min_max(x, self.min_file_amount, self.max_file_amount, 'File amount')

    def _check_min_max(self, size: int, min_size: int | None, max_size: int | None, part: str) -> None:
        if min_size is not None and size < min_size:
            raise ActionError(f'Received {part} is less than allowed [{min_size}]', action=self.action)
        if max_size is not None and max_size < size:
            raise ActionError(f'Received {part} is more than allowed [{max_size}]', action=self.action)



def main(host: str, port: int,
         handshake: bytes | None,
         api: int,
         input_timeout: float,
         idle_timeout: float,
         tls: bool,
         debug: bool):
    import logging
    from tornado.ioloop import IOLoop
    from cats.v2 import Config, SHA256TimeHandshake
    from cats.v2.server import Api, Application, Handler, Server

    logging.basicConfig(level='DEBUG' if debug else 'INFO', force=True)

    api = Api()

    class EchoHandler(Handler, api=api, id=0xFFFF):  # noqa
        async def handle(self):
            return self.action

    if handshake is not None:
        handshake = SHA256TimeHandshake(secret_key=handshake, valid_window=1, timeout=5.0)
    config = Config(
        idle_timeout=idle_timeout,
        input_timeout=input_timeout,
        input_limit=5,
        debug=debug,
        handshake=handshake
    )
    ssl_options = None
    if tls:
        from ssl import SSLContext, PROTOCOL_TLS_CLIENT, CERT_REQUIRED
        ssl_options = SSLContext(protocol=PROTOCOL_TLS_CLIENT)
        ssl_options.verify_mode = CERT_REQUIRED
        ssl_options.load_default_certs()

    app = Application([api], config=config)
    cats_server = Server(app, ssl_options=ssl_options)
    cats_server.bind(port, host)
    cats_server.start(1)
    IOLoop.current().start()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--host', '-H', dest='host', type=str,
                        default='0.0.0.0', help='CATS.v2.server host. Default: 0.0.0.0')
    parser.add_argument('--port', '-P', dest='port', type=int,
                        default=9191, help='CATS.v2.server port. Default: 9191')
    parser.add_argument('--handshake', dest='handshake', type=bytes,
                        default=None, help='Handshake secret key. Default: [disabled]')
    parser.add_argument('--input-timeout', dest='input_timeout', type=float,
                        default=120.0, help='Input timeout. Default: 120.0')
    parser.add_argument('--api', '-A', dest='api', type=int,
                        default=1, help='API version. Default: 1')
    parser.add_argument('--idle-timeout', dest='idle_timeout', type=float,
                        default=120.0, help='Idle timeout. Default: 120.0')
    parser.add_argument('--tls', '-T', dest='tls', action='store_true',
                        default=False, help='Enable TLS. Default: False')
    parser.add_argument('--debug', '-D', dest='debug', action='store_true',
                        default=False, help='Enable debug. Default: False')
    args = vars(parser.parse_args())
    try:
        main(**args)
    except KeyboardInterrupt:
        print()



from enum import IntEnum
from typing import TypeAlias

import msgpack
import toml
import ujson
import yaml

__all__ = [
    'Data',
    'SchemeType',
    'Scheme',
    'JSON',
    'YAML',
    'TOML',
    'MsgPack',
]

Data: TypeAlias = int | float | bool | str | None
Data: TypeAlias = Data | list[Data] | dict[str, Data]


class SchemeType(IntEnum):
    json = 1
    yaml = 2
    toml = 3
    msgpack = 4


class Scheme:
    id: SchemeType

    @classmethod
    def loads(cls, buff: bytes | str) -> Data:
        """
        Parse byte/str buffer to data (primitives or list/dict of primitives)
        :param buff: Encoded buffer
        :type buff: bytes | str
        :return: Decoded buffer
        """
        raise NotImplementedError

    @classmethod
    def dumps(cls, data: Data) -> bytes:
        """
        Encode data into byte buffer
        :param data: primitives or list/dict of primitives
        :type data: Data
        :return: Encoded buffer
        """
        raise NotImplementedError


class JSON(Scheme):
    id = SchemeType.json

    @classmethod
    def loads(cls, data: bytes | str) -> Data:
        return ujson.loads(data)

    @classmethod
    def dumps(cls, data: Data) -> bytes:
        return ujson.dumps(data).encode('utf-8')


class YAML(Scheme):
    id = SchemeType.yaml

    @classmethod
    def loads(cls, buff: bytes | str) -> Data:
        return yaml.safe_load(buff)

    @classmethod
    def dumps(cls, data: Data) -> bytes:
        return yaml.safe_dump(data, encoding='utf-8')


class TOML(Scheme):
    id = SchemeType.toml

    @classmethod
    def loads(cls, buff: bytes | str) -> Data:
        return toml.loads(buff)

    @classmethod
    def dumps(cls, data: Data) -> bytes:
        return toml.dumps(data).encode('utf-8')


class MsgPack(Scheme):
    id = SchemeType.msgpack

    @classmethod
    def loads(cls, buff: bytes) -> Data:
        return msgpack.loads(buff)

    @classmethod
    def dumps(cls, data: Data) -> bytes:
        return msgpack.dumps(data)



from dataclasses import asdict, dataclass
from typing import Literal, TypeVar

import ujson

from cats import to_uint

__all__ = [
    'ClientStatement',
    'ServerStatement',
]
T = TypeVar('T', bound='Statement')


@dataclass
class Statement:
    def pack(self) -> bytes:
        data: str = ujson.dumps(asdict(self), ensure_ascii=False, escape_forward_slashes=False)
        return to_uint(len(data), 4) + data.encode('utf-8')

    @classmethod
    def unpack(cls, buffer: bytes) -> T:
        return cls(**ujson.loads(buffer))  # noqa


@dataclass
class ClientStatement(Statement):
    api: int
    client_time: int
    scheme_format: Literal['JSON']
    compressors: list[Literal['gzip', 'zlib']]
    default_compression: Literal['gzip', 'zlib']


@dataclass
class ServerStatement(Statement):
    server_time: int



from dataclasses import dataclass
from io import BytesIO
from os.path import getsize
from pathlib import Path
from typing import IO, TypeAlias

import ujson

from cats.errors import *
from cats.plugins import BaseModel, BaseSerializer, Form, ModelSchema, scheme_json
from cats.types import Byte, Json, List, T_Headers
from cats.utils import tmp_file

__all__ = [
    'FileInfo',
    'Files',
    'BaseCodec',
    'ByteCodec',
    'JsonCodec',
    'FileCodec',
    'T_BYTE',
    'T_JSON',
    'T_FILE',
    'Codec',
]


@dataclass
class FileInfo:
    name: str
    path: Path
    size: int
    mime: str | None


class Files(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in self.items():
            if not isinstance(k, str):
                raise ValidationError('File key must be string', data=self)
            elif not isinstance(v, FileInfo):
                raise ValidationError('File value must be FileInfo', data=self)

    def __del__(self):
        for v in self.values():
            v.path.unlink(missing_ok=True)


class BaseCodec:
    type_id: int
    type_name: str

    @classmethod
    async def encode(cls, data, headers: T_Headers) -> bytes:
        """
        :raise TypeError: Encoder doesn't support this type
        :raise ValueError: Failed to encode
        """
        raise NotImplementedError

    @classmethod
    async def decode(cls, data: bytes | Path, headers: T_Headers):
        """
        :raise TypeError: Encoder doesn't support this type
        :raise ValueError: Failed to decode
        """
        raise NotImplementedError


class ByteCodec(BaseCodec):
    type_id = 0x00
    type_name = 'bytes'

    @classmethod
    async def encode(cls, data: Byte, headers: T_Headers, offset: int = 0) -> bytes:
        if data is not None and not isinstance(data, Byte):
            raise InvalidCodecError(f'{cls} does not support {type(data)}', data=data, headers=headers)

        return (bytes(data) if data else bytes())[offset:]

    @classmethod
    async def decode(cls, data: bytes, headers: T_Headers) -> bytes:
        return bytes(data) if data else bytes()


class JsonCodec(BaseCodec):
    type_id = 0x01
    type_name = 'json'

    @classmethod
    async def encode(cls, data: Json | Form | list[Form], headers: T_Headers, offset: int = 0) -> bytes:
        try:
            if data:
                if isinstance(data, (BaseModel, BaseSerializer, ModelSchema)):
                    return scheme_json(type(data), data, many=False, plain=True)
                elif isinstance(data, List):
                    data = list(data)
                    if isinstance(data[0], (BaseModel, BaseSerializer, ModelSchema)):
                        return scheme_json(type(data[0]), data, many=True, plain=True)
            return await cls._encode(data, offset=offset)
        except TypeError as err:
            raise InvalidCodecError(f'{cls} does not support {type(data)}', data=data, headers=headers) from err

    @classmethod
    async def _encode(cls, data: Json, offset: int = 0) -> bytes:
        if not isinstance(data, (str, int, float, dict, list, bool, type(None))):
            raise TypeError

        return ujson.dumps(data, ensure_ascii=False, escape_forward_slashes=False).encode('utf-8')[offset:]

    @classmethod
    async def decode(cls, data: bytes, headers) -> Json:
        if not data:
            return {}

        try:
            return ujson.loads(data.decode('utf-8'))
        except ValueError as err:
            raise CodecError('Failed to parse JSON from data', data=data, headers=headers) from err


FILE_TYPES: TypeAlias = Path | list[Path] | dict[str, Path] | FileInfo | list[FileInfo] | dict[str, FileInfo]


class FileCodec(BaseCodec):
    type_id = 0x02
    type_name = 'files'
    encoding = 'utf-8'

    @classmethod
    def path_to_file_info(cls, path: Path | FileInfo) -> FileInfo:
        if isinstance(path, FileInfo):
            return path
        elif isinstance(path, Path):
            return FileInfo(path.name, path, getsize(path.as_posix()), None)
        raise TypeError

    @classmethod
    def normalize_input(cls, data: FILE_TYPES) -> dict[str, FileInfo]:
        if isinstance(data, Path):
            data = cls.path_to_file_info(data)
        elif isinstance(data, list):
            res = []
            for i in data:
                res.append(cls.path_to_file_info(i))

        elif isinstance(data, dict):
            res = {}
            for k, v in data.items():
                if not isinstance(k, str):
                    raise TypeError
                res[k] = cls.path_to_file_info(v)
        elif not isinstance(data, FileInfo):
            raise TypeError

        if isinstance(data, FileInfo):
            return {data.name: data}
        elif isinstance(data, list):
            return {i.name: i for i in data}
        else:
            return data

    @classmethod
    async def encode(cls, data: FILE_TYPES, headers: T_Headers, offset: int = 0) -> Path:
        tmp = tmp_file()
        try:
            buff = tuple(cls.normalize_input(data).items())
            header = []
            with tmp.open('wb') as fh:
                for key, info in buff:
                    left = info.size - offset
                    offset = max(0, offset - left)
                    if left < 0:
                        continue
                    header.append({"key": key, "name": info.name, "size": left, "type": info.mime})
                    with info.path.open('rb') as f_fh:
                        f_fh.seek(info.size - left)
                        while left > 0:
                            buff = f_fh.read(1 << 24)
                            if not len(buff):
                                raise ValueError
                            fh.write(buff)
                            left -= len(buff)
            headers['Files'] = header
            return tmp

        except (KeyError, ValueError, TypeError, AttributeError) as err:
            tmp.unlink(missing_ok=True)
            raise InvalidCodecError(f'{cls} does not support {type(data)}', data=data, headers=headers) from err

    @classmethod
    async def decode(cls, data: Path | bytes | bytearray, headers) -> Files:
        result = Files()
        buff = data.open('rb') if isinstance(data, Path) else BytesIO(data)

        try:
            if 'Files' not in headers:
                raise MalformedDataError('Failed to parse Files meta data from headers', data=data, headers=headers)

            header = headers['Files']

            if not isinstance(header, list):
                raise MalformedDataError(f'Files headers must be list, {type(header)} provided',
                                         data=data, headers=headers)

            for i, node in enumerate(header):
                if not isinstance(node, dict):
                    raise MalformedDataError(f'Files header item[{i}] must be dict, {type(node)} provided',
                                             data=data, headers=headers)

                tmp = await cls._unpack_file(buff, node)
                result[node['key']] = FileInfo(
                    name=node['name'],
                    path=tmp,
                    size=node['size'],
                    mime=node.get('type'),
                )

            return result
        except (KeyError, ValueError, TypeError) as err:
            for v in result.values():
                v.path.unlink(missing_ok=True)
            raise CodecError('Failed to parse Files from data', data=data, headers=headers) from err
        finally:
            buff.close()

    @classmethod
    async def _unpack_file(cls, fh: IO, node) -> Path:
        tmp = tmp_file()
        left = node['size']
        with tmp.open('wb') as node_fh:
            while left > 0:
                buff = fh.read(min(left, 1 << 24))
                if not len(buff):
                    raise ValueError('Failed to unpack file: payload size exceeded')
                node_fh.write(buff)
                left -= len(buff)
        return tmp


T_BYTE = ByteCodec.type_id
T_JSON = JsonCodec.type_id
T_FILE = FileCodec.type_id


class Codec:
    codecs = {
        T_BYTE: ByteCodec,
        T_JSON: JsonCodec,
        T_FILE: FileCodec,
    }

    @classmethod
    async def encode(cls, buff: Byte | Json | FILE_TYPES, headers: T_Headers, offset: int = 0) -> (bytes, int):
        """
        Takes any supported data type and returns tuple (encoded: bytes, type_id: int)
        """
        for type_id, codec in cls.codecs.items():
            try:
                encoded = await codec.encode(buff, headers, offset)
                return encoded, type_id
            except InvalidCodecError:
                continue

        raise CodecError(f'Failed to encode data: Type {type(buff)} not supported', data=buff, headers=headers)

    @classmethod
    async def decode(cls, buff: Byte | Path, data_type: int, headers: T_Headers) -> Byte | Json | Files:
        """
        Takes byte buffer, type_id and try to decode it to internal data types
        """
        if data_type not in cls.codecs:
            raise CodecError(f'Failed to decode data: Type {data_type} not supported', data=buff, headers=headers)

        return await cls.codecs[data_type].decode(buff, headers)

    def get_codec_name(self, type_id: int, default: str = 'unknown') -> str:
        """
        Returns Type Name by it's id (w/ fallback to default)
        :param type_id:
        :param default:
        :return:
        """
        try:
            return self.codecs[type_id].type_name
        except KeyError:
            return default



from typing import Type, TypeAlias, TypeVar

import ujson

from cats.errors import UnsupportedSchemeError
from cats.types import Json, Missing, Model, QuerySet

try:
    from rest_framework.serializers import BaseSerializer
except ImportError:
    BaseSerializer = type('BaseSerializer', (object,), {})

try:
    from pydantic import BaseModel
except ImportError:
    BaseModel = type('BaseModel', (object,), {})

try:
    from djantic import ModelSchema
except ImportError:
    ModelSchema = type('ModelSchema', (object,), {})

__all__ = [
    'QuerySet',
    'Model',
    'BaseSerializer',
    'BaseModel',
    'ModelSchema',
    'Scheme',
    'SchemeTypes',
    'Form',
    'without_missing',
    'scheme_load',
    'scheme_dump',
    'scheme_json',
]

Scheme: TypeAlias = Type[BaseModel] | Type[BaseSerializer] | Type[ModelSchema]
SchemeTypes = BaseModel, BaseSerializer, ModelSchema
PydanticModel = TypeVar('PydanticModel', bound=BaseModel)
DRFModel = TypeVar('DRFModel', bound=BaseSerializer)
DjanticModel = TypeVar('DjanticModel', bound=ModelSchema)
Form: TypeAlias = PydanticModel | DRFModel | DjanticModel


def without_missing(obj) -> Json | None:
    if _items := getattr(obj, 'items', None):
        return {k: without_missing(v) for k, v in _items() if not isinstance(v, Missing)}
    if isinstance(obj, (list, set, tuple)):
        return [without_missing(i) for i in obj if not isinstance(i, Missing)]
    if isinstance(obj, Missing):
        return None
    return obj


class DRF:
    @classmethod
    def load(cls, s: Type[BaseSerializer], data: Json, *, many: bool = False, plain: bool = False) -> Json | DRFModel:
        f = s(data=data, many=many)
        f.is_valid(raise_exception=True)
        if plain:
            return f
        return without_missing(f.validated_data)

    @classmethod
    def dump(cls, s: Type[BaseSerializer], data: Json | DRFModel, *, many: bool = False, plain: bool = False) -> Json:
        if plain:
            assert isinstance(data, s)
            return data.data
        return s(data, many=many).data

    @classmethod
    def json(cls, s: Type[BaseSerializer], data: Json | DRFModel, *, many: bool = False, plain: bool = False) -> bytes:
        obj = cls.dump(s, data, many=many, plain=plain)
        dump: str = ujson.dumps(obj, ensure_ascii=False, escape_forward_slashes=False)
        return dump.encode('utf-8')


class Pydantic:
    @classmethod
    def load(cls, s: Type[BaseModel], data: Json, *, many: bool = False, plain: bool = False) -> Json | PydanticModel:
        if many:
            return [cls.load(s, i, many=False, plain=plain) for i in data]

        res = s.parse_obj(data)
        if plain:
            return res

        return without_missing(res.dict())

    @classmethod
    def dump(cls, s: Type[BaseModel], data: Json | PydanticModel, *, many: bool = False, plain: bool = False) -> Json:
        if many:
            return [cls.dump(s, i, many=False, plain=plain) for i in data]

        if plain:
            assert isinstance(data, s)
        elif isinstance(data, (QuerySet, Model)):
            data = s.from_orm(data)
        else:
            data = s.parse_obj(data)
        return data.dict()

    @classmethod
    def json(cls, s: Type[BaseModel], data: Json | PydanticModel, *, many: bool = False, plain: bool = False) -> bytes:
        if many:
            return b'[' + b','.join(cls.json(s, i, many=False, plain=plain) for i in data) + b']'

        if plain:
            assert isinstance(data, s)
        elif isinstance(data, (QuerySet, Model)):
            data = s.from_orm(data)
        else:
            data = s.parse_obj(data)
        res = data.json()
        if isinstance(res, bytes):
            return res
        return res.encode('utf-8')


class Djantic(Pydantic):
    @classmethod
    def dump(cls, s: Type[ModelSchema], data: Json, *, many: bool = False, plain: bool = False) -> Json | DjanticModel:
        if many:
            return [cls.dump(s, i, many=False, plain=plain) for i in data]

        if not isinstance(data, (QuerySet, Model)):
            return super().dump(s, data, many=many, plain=plain)

        if plain:
            assert isinstance(data, s)
        else:
            data = s.from_django(data)
        return data.json()

    @classmethod
    def json(cls, s: Type[ModelSchema], data: Json | DjanticModel, *, many: bool = False, plain: bool = False) -> bytes:
        if many:
            return b'[' + b','.join(cls.json(s, i, many=False, plain=plain) for i in data) + b']'

        if not isinstance(data, (QuerySet, Model)):
            return super().json(s, data, many=many, plain=plain)

        if plain:
            assert isinstance(data, s)
        else:
            data = s.from_django(data)
        res = data.json()
        if isinstance(res, bytes):
            return res
        return res.encode('utf-8')


def _resolve_scheme_type(scheme: Scheme) -> Type[DRF] | Type[Djantic] | Type[Pydantic]:
    if issubclass(scheme, BaseSerializer):
        return DRF
    elif issubclass(scheme, ModelSchema):
        return Djantic
    elif issubclass(scheme, BaseModel):
        return Pydantic
    else:
        raise UnsupportedSchemeError('Unsupported scheme', scheme=scheme)


def scheme_load(scheme: Scheme, data: Json, *, many: bool = False, plain: bool = False) -> Json | Form:
    return _resolve_scheme_type(scheme).load(scheme, data, many=many, plain=plain)


def scheme_dump(scheme: Scheme, data: Json | Form, *, many: bool = False, plain: bool = False) -> Json:
    return _resolve_scheme_type(scheme).dump(scheme, data, many=many, plain=plain)


def scheme_json(scheme: Scheme, data: Json | Form, *, many: bool = False, plain: bool = False) -> bytes:
    return _resolve_scheme_type(scheme).json(scheme, data, many=many, plain=plain)



from collections import defaultdict
from functools import partial
from typing import Awaitable, Iterable, Type

from cats.v2.action import ActionLike
from cats.v2.auth import Auth
from cats.v2.config import Config
from cats.v2.connection import Connection
from cats.v2.server.handlers import Api, Handler, HandlerItem
from cats.v2.server.middleware import Forward, Middleware, default_error_handler

__all__ = [
    'Application',
]


class Application:
    __slots__ = ('config', 'auth', 'ConnectionClass', '_handlers', '_channels', '_runner')

    def __init__(self, apis: list[Api], middleware: list[Middleware] = None, *,
                 auth: Auth = None, config: Config = None, connection: Type[Connection] = None):
        self.config = Config() if config is None else config
        self.auth = Auth([]) if auth is None else auth
        self.ConnectionClass = connection
        if middleware is None:
            middleware = [
                default_error_handler,
            ]
        self._channels: dict[str, list[Connection]] = defaultdict(list)

        api = Api()
        for i in apis:
            api.update(i)

        self._handlers = api.compute()
        self._runner: Forward = self._run
        if middleware:
            for md in middleware:
                self._runner = partial(md, forward=self._runner)

    async def _run(self, handler: Handler) -> ActionLike | None:
        return await handler()

    def run(self, handler: Handler) -> Awaitable[ActionLike | None]:
        return self._runner(handler)

    def get_handlers_by_id(self, handler_id: int) -> list[HandlerItem] | HandlerItem | None:
        return self._handlers.get(handler_id)

    def get_handler_id(self, handler: Handler) -> int | None:
        return handler.handler_id

    def channels(self) -> list[str]:
        return list(self._channels.keys())

    def channel(self, name: str) -> Iterable[Connection]:
        return iter(self._channels.get(name, []))

    def attach_conn_to_channel(self, conn: Connection, channel: str) -> None:
        if conn not in self._channels[channel]:
            self._channels[channel].append(conn)

    def detach_conn_from_channel(self, conn: Connection, channel: str) -> None:
        try:
            self._channels[channel].remove(conn)
        except (KeyError, ValueError):
            pass

    def clear_channel(self, channel: str) -> None:
        self._channels[channel].clear()

    def clear_all_channels(self) -> None:
        self._channels.clear()

    def remove_conn_from_channels(self, conn: Connection) -> None:
        for channel_name in self.channels():
            self.detach_conn_from_channel(conn, channel_name)



"""
@dataclass
class Config:
    idle_timeout: float | int = 120.0
    input_timeout: float | int = 120.0
    input_limit: int = 5
    debug: bool = False
    max_plain_payload: int = 16 * 1024 * 1024
    stream_errors: Type[Exception] | tuple[Type[Exception]] = (
        asyncio.TimeoutError,
        asyncio.CancelledError,
        asyncio.InvalidStateError,
        StreamClosedError,
    )
    ignore_errors: Type[Exception] | tuple[Type[Exception]] = (
        *stream_errors,
        CatsError,
        KeyboardInterrupt,
    )
    handshake: Handshake | None = None

class Options:
    api_version: int = 1
    available_compressors = [GZip, zLib]
    default_compressor = none | GZip | zLib
    scheme_type: json | yaml | xml = json
    download_speed: int | None

class User:
    identity: Identity
    credentials: Credentials
    timer: Timer

class Connection:
    config: Config
    options: Options

class Codec[type_id]:
    def dispatch(data, headers, options: Options): most fit codec
    input = [data, headers]
    output = [data, headers]

class Compressor[type_id]:
    def dispatch(data, headers, options: Options): most fit compressor
    input = [data, headers]
    output = [data, headers]
"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cats',
 'cats.test_utils',
 'cats.test_utils.client',
 'cats.test_utils.pytest',
 'cats.v2',
 'cats.v2.client',
 'cats.v2.server',
 'cats.v3']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'msgpack>=1.0.3,<2.0.0',
 'pytz>=2021.3,<2022.0',
 'richerr>=0.2.3,<0.3.0',
 'rollbar>=0.16.2,<0.17.0',
 'struct-model-python>=0.1.1,<0.2.0',
 'toml>=0.10.2,<0.11.0',
 'tornado>=6.1,<7.0']

extras_require = \
{'django': ['Django>=2.0,<3.0', 'djangorestframework>=3.12.4'],
 'djantic': ['Django>=2.0,<3.0', 'pydantic>=1.9.0', 'djantic>=0.3.3'],
 'pydantic': ['pydantic>=1.9.0']}

entry_points = \
{'pytest11': ['pytest_cats = cats.test_utils.pytest.plugin']}

setup_kwargs = {
    'name': 'cats-python',
    'version': '4.5.11',
    'description': 'Cifrazia Action Transport System for Python',
    'long_description': '# Welcome\n\n[![PyPI version](https://badge.fury.io/py/cats-python.svg)](https://badge.fury.io/py/cats-python) [![codecov](https://codecov.io/gh/Cifrazia/cats-python/branch/master/graph/badge.svg?token=MMDPS40REC)](https://codecov.io/gh/Cifrazia/cats-python) [![FOSSA Status](https://app.fossa.com/api/projects/git%2Bgithub.com%2FCifrazia%2Fcats-python.svg?type=shield)](https://app.fossa.com/projects/git%2Bgithub.com%2FCifrazia%2Fcats-python?ref=badge_shield)\n\n## Cifrazia Action Transport System\n\nCATS - is a TCP based byte protocol for persistence package exchanging. This so-called protocol is designed specifically\nfor internal use in [Cifrazia](https://cifrazia.com).\n\n[Learn more about protocol](./protocol)\n\n## Features\n\n+ One action at a time\n+ Up-to 4GB payload in single [plain Action](protocol/2.0.md#0x00-action)\n+ Unlimited and delayed payload in [Streaming Actions](protocol/2.0.md#0x01-streamaction)\n+ Chained [inputs](protocol/2.0.md#inputs)\n+ [Broadcasts](protocol/2.0.md#broadcast)\n+ Multiple [data formats](protocol/2.0.md#data-types)\n+ Custom [handshakes](protocol/2.0.md#handshake)\n+ ~~Local and global encryption~~\n\n[!ref](get-started.md)\n\n## Requirements\n\n+ Python `^3.9`\n+ Tornado `^6.1`\n+ Sentry SDK `^1.1.0`\n+ uJSON `^4.0.2`\n+ PyTZ `^2021.1`\n\n[![FOSSA Status](https://app.fossa.com/api/projects/git%2Bgithub.com%2FCifrazia%2Fcats-python.svg?type=large)](https://app.fossa.com/projects/git%2Bgithub.com%2FCifrazia%2Fcats-python?ref=badge_large)',
    'author': 'Bogdan Parfenov',
    'author_email': 'adam.brian.bright@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://adambrianbright.github.io/cats-python/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10.0,<4.0',
}


setup(**setup_kwargs)



import asyncio
import hashlib
from time import time

from cats.errors import HandshakeError
from cats.utils import bytes2hex

__all__ = [
    'Handshake',
    'SHA256TimeHandshake',
]


class Handshake:
    async def validate(self, conn) -> None:
        """
        Read stream, check for valid handshake
        If valid: send 0x01 to stream
        Otherwise: send 0x00 to stream and close
        :param conn:
        :return:
        """
        raise NotImplementedError

    async def send(self, conn) -> None:
        """
        Send valid handshake to stream and read answer
        If 0x01 - passing
        Otherwise: raising HandshakeError
        :param conn:
        :return:
        """
        raise NotImplementedError


class SHA256TimeHandshake(Handshake):
    __slots__ = ('secret_key', 'valid_window', 'timeout')

    def __init__(self, secret_key: bytes, valid_window: int = None, timeout: int | float = 5.0):
        assert isinstance(secret_key, bytes) and secret_key
        self.secret_key = secret_key
        self.valid_window = valid_window or 1
        self.timeout = timeout
        assert self.valid_window >= 1

    def get_hashes(self, timestamp: float) -> list[bytes]:
        ts = round(timestamp / 10) * 10

        return [
            hashlib.sha256(self.secret_key + str(ts + i * 10).encode('utf-8')).digest()
            for i in range(-self.valid_window, self.valid_window + 1)
        ]

    async def validate(self, conn) -> None:
        handshake: bytes = await asyncio.wait_for(conn.read(32), self.timeout)
        conn.debug(f'[RECV {conn.address}] Handshake: {bytes2hex(handshake)}')
        try:
            if handshake not in self.get_hashes(time()):
                await conn.write(b'\x00')
                conn.debug(f'[SEND {conn.address} Handshake failed')
                raise HandshakeError('Invalid handshake', conn=conn, handshake=handshake)
        except UnicodeDecodeError as err:
            await conn.write(b'\x00')
            raise HandshakeError('Malformed handshake', conn=conn, handshake=handshake) from err
        else:
            await conn.write(b'\x01')
            conn.debug(f'[SEND {conn.address}] Handshake passed')

    async def send(self, conn) -> None:
        handshake = self.get_hashes(conn.time_delta + time())[1]
        await conn.write(handshake)
        result = await conn.read(1)
        if result == b'\x01':
            conn.debug(f'[SEND {conn.address}] Handshake passed')
            return
        raise HandshakeError('Invalid handshake', conn=conn, handshake=handshake)



from tornado.iostream import StreamClosedError

from cats.errors import HandshakeError, ProtocolError
from cats.v2.action import PingAction


async def main(
    host: str, port: int,
    handshake: bytes | None,
    api: int,
    input_timeout: float,
    idle_timeout: float,
    tls: bool,
    debug: bool
):
    import logging
    from cats.v2 import Config, SHA256TimeHandshake
    from cats.v2.client import Connection

    logging.basicConfig(level='DEBUG' if debug else 'INFO', force=True)

    if handshake is not None:
        handshake = SHA256TimeHandshake(secret_key=handshake, valid_window=1, timeout=5.0)
    config = Config(
        idle_timeout=idle_timeout,
        input_timeout=input_timeout,
        input_limit=5,
        debug=debug,
        handshake=handshake,
    )
    conn = Connection(config, api_version=api)
    ssl_options = None
    if tls:
        from ssl import SSLContext, PROTOCOL_TLS_CLIENT, CERT_REQUIRED
        ssl_options = SSLContext(protocol=PROTOCOL_TLS_CLIENT)
        ssl_options.verify_mode = CERT_REQUIRED
        ssl_options.load_default_certs()

    try:
        await conn.connect(host, port, ssl_options=ssl_options)
        await PingAction().send(conn)
        exit(int(not conn.is_open))
    except (HandshakeError, ProtocolError):
        exit(0)
    except StreamClosedError:
        exit(1)
    except Exception:
        exit(500)
        raise


if __name__ == '__main__':
    import asyncio
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--host', '-H', dest='host', type=str,
        default='127.0.0.1', help='CATS.v2.server host. Default: 127.0.0.1',
    )
    parser.add_argument(
        '--port', '-P', dest='port', type=int,
        default=9191, help='CATS.v2.server port. Default: 9191',
    )
    parser.add_argument(
        '--handshake', dest='handshake', type=bytes,
        default=None, help='Handshake secret key. Default: [disabled]',
    )
    parser.add_argument(
        '--input-timeout', dest='input_timeout', type=float,
        default=120.0, help='Server`s input timeout. Default: 120.0',
    )
    parser.add_argument(
        '--api', '-A', dest='api', type=int,
        default=1, help='API version. Default: 1',
    )
    parser.add_argument(
        '--idle-timeout', dest='idle_timeout', type=float,
        default=120.0, help='Server`s idle timeout. Default: 120.0',
    )
    parser.add_argument(
        '--tls', '-T', dest='tls', action='store_true',
        default=False, help='Enable TLS. Default: False',
    )
    parser.add_argument(
        '--debug', '-D', dest='debug', action='store_true',
        default=False, help='Enable debug. Default: False',
    )
    args = vars(parser.parse_args())
    asyncio.run(main(**args))



from typing import Type, TypeVar

IdentityObject = TypeVar('IdentityObject', bound='Identity')


class Identity:
    __slots__ = ()
    identity_list: list[Type['Identity']] = []

    @property
    def id(self) -> int:
        raise NotImplementedError

    @property
    def model_name(self) -> int:
        raise NotImplementedError

    # Rename to excerpt -> dict[str]
    @property
    def sentry_scope(self) -> dict:
        raise NotImplementedError
